#!/usr/bin/env python
###############################################################################
#                                                                             #
#    mstore.py                                                                #
#                                                                             #
#    GroopM - Low level data management and file parsing                      #
#                                                                             #
#    Copyright (C) Michael Imelfort                                           #
#                                                                             #
###############################################################################
#                                                                             #
#          .d8888b.                                    888b     d888          #
#         d88P  Y88b                                   8888b   d8888          #
#         888    888                                   88888b.d88888          #
#         888        888d888 .d88b.   .d88b.  88888b.  888Y88888P888          #
#         888  88888 888P"  d88""88b d88""88b 888 "88b 888 Y888P 888          #
#         888    888 888    888  888 888  888 888  888 888  Y8P  888          #
#         Y88b  d88P 888    Y88..88P Y88..88P 888 d88P 888   "   888          #
#          "Y8888P88 888     "Y88P"   "Y88P"  88888P"  888       888          #
#                                             888                             #
#                                             888                             #
#                                             888                             #
#                                                                             #
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = "Michael Imelfort"
__copyright__ = "Copyright 2012/2013"
__credits__ = ["Michael Imelfort"]
__license__ = "GPL3"
__version__ = "0.3.6"
__maintainer__ = "Michael Imelfort"
__email__ = "mike@mikeimelfort.com"
__status__ = "Release"
__current_GMDB_version__ = 5

###############################################################################

from sys import exc_info
from os.path import splitext as op_splitext, basename as op_basename
from string import maketrans as s_maketrans

import tables
import numpy as np
from scipy.spatial.distance import cdist, squareform
import pysam

# GroopM imports
from PCA import PCA, Center

np.seterr(all='raise')

# shut up pytables!
import warnings
warnings.filterwarnings('ignore', category=tables.NaturalNameWarning)

###############################################################################
###############################################################################
###############################################################################
###############################################################################
class GMDataManager:
    """Top level class for manipulating GroopM data

    Use this class for parsing in raw data into a hdf DB and
    for reading from and updating same DB

    NOTE: All tables are kept in the same order indexed by the contig ID
    Tables managed by this class are listed below

    ------------------------
     PROFILES
    group = '/profile'
    ------------------------
    **Kmer Signature**
    table = 'kms'
    'mer1' : tables.FloatCol(pos=0)
    'mer2' : tables.FloatCol(pos=1)
    'mer3' : tables.FloatCol(pos=2)
    ...

    **Kmer Vals**
    table = 'kpca'
    'pc1' : tables.FloatCol(pos=0)
    'pc2' : tables.FloatCol(pos=1)
    'pc3' : tables.FloatCol(pos=2)
    ...

    **Coverage profile**
    table = 'coverage'
    'stoit1' : tables.FloatCol(pos=0)
    'stoit2' : tables.FloatCol(pos=1)
    'stoit3' : tables.FloatCol(pos=2)
    ...

    **Transformed coverage profile**
    table = 'transCoverage'
    'x' : tables.FloatCol(pos=0)
    'y' : tables.FloatCol(pos=1)
    'z' : tables.FloatCol(pos=2)

    **Normalised coverage profile**
    table = 'normCoverage'
    'normCov' : tables.FloatCol(pos=0)

    ------------------------
     LINKS
    group = '/links'
    ------------------------
    ** Links **
    table = 'links'
    'contig1'    : tables.Int32Col(pos=0)            # reference to index in meta/contigs
    'contig2'    : tables.Int32Col(pos=1)            # reference to index in meta/contigs
    'numReads'   : tables.Int32Col(pos=2)            # number of reads supporting this link
    'linkType'   : tables.Int32Col(pos=3)            # the type of the link (SS, SE, ES, EE)
    'gap'        : tables.Int32Col(pos=4)            # the estimated gap between the contigs

    ------------------------
     METADATA
    group = '/meta'
    ------------------------
    ** Metadata **
    table = 'meta'
    'stoitColNames' : tables.StringCol(512, pos=0)
    'numStoits'     : tables.Int32Col(pos=1)
    'merColNames'   : tables.StringCol(4096,pos=2)
    'merSize'       : tables.Int32Col(pos=3)
    'numMers'       : tables.Int32Col(pos=4)
    'numCons'       : tables.Int32Col(pos=5)
    'numBins'       : tables.Int32Col(pos=6)
    'clustered'     : tables.BoolCol(pos=7)           # set to true after clustering is complete
    'complete'      : tables.BoolCol(pos=8)           # set to true after clustering finishing is complete
    'formatVersion' : tables.Int32Col(pos=9)          # groopm file version

    **PC variance**
    table = 'kpca_variance'
    'pc1_var' : tables.FloatCol(pos=0)
    'pc2_var' : tables.FloatCol(pos=1)
    'pc3_var' : tables.FloatCol(pos=2)
    ...

    ** Contigs **
    table = 'contigs'
    'cid'    : tables.StringCol(512, pos=0)
    'bid'    : tables.Int32Col(pos=1)
    'length' : tables.Int32Col(pos=2)
    'gc'     : tables.FloatCol(pos=3)

    ** Bins **
    table = 'bins'
    'bid'        : tables.Int32Col(pos=0)
    'numMembers' : tables.Int32Col(pos=1)
    'isLikelyChimeric' : tables.BoolCol(pos=2)

    **Transformed coverage corners**
    table = 'transCoverageCorners'
    'x' : tables.FloatCol(pos=0)
    'y' : tables.FloatCol(pos=1)
    'z' : tables.FloatCol(pos=2)

    """
    def __init__(self): pass

#------------------------------------------------------------------------------
# DB CREATION / INITIALISATION  - PROFILES

    def createDB(self, bamFiles, contigs, dbFileName, timer, kmerSize=4, force=False):
        """Main wrapper for parsing all input files"""
        # load all the passed vars
        dbFileName = dbFileName
        contigsFile = contigs
        stoitColNames = []

        kse = KmerSigEngine(kmerSize)
        conParser = ContigParser()
        bamParser = BamParser()

        cid_2_indices = {}

        # make sure we're only overwriting existing DBs with the users consent
        try:
            with open(dbFileName) as f:
                if(not force):
                    user_option = self.promptOnOverwrite(dbFileName)
                    if(user_option != "Y"):
                        print "Operation cancelled"
                        return False
                    else:
                        print "Overwriting database",dbFileName
        except IOError as e:
            print "Creating new database", dbFileName

        # create the db
        try:
            with tables.openFile(dbFileName, mode = "w", title = "GroopM") as h5file:
                # Create groups under "/" (root) for storing profile information and metadata
                profile_group = h5file.createGroup("/", 'profile', 'Assembly profiles')
                meta_group = h5file.createGroup("/", 'meta', 'Associated metadata')
                links_group = h5file.createGroup("/", 'links', 'Paired read link information')
                #------------------------
                # parse contigs
                #
                # Contig IDs are key. Any keys existing in other files but not in this file will be
                # ignored. Any missing keys in other files will be given the default profile value
                # (typically 0). Ironically, we don't store the CIDs here, these are saved one time
                # only in the bin table
                #
                # Before writing to the database we need to make sure that none of them have
                # 0 coverage @ all stoits.
                #------------------------
                try:
                    with open(contigsFile, "r") as f:
                        try:
                            (con_names, con_gcs, con_lengths, con_ksigs) = conParser.parse(f, kse)
                            num_cons = len(con_names)
                            cid_2_indices = dict(zip(con_names, range(num_cons)))
                        except:
                            print "Error parsing contigs"
                            raise
                except:
                    print "Could not parse contig file:",contigsFile,exc_info()[0]
                    raise

                #------------------------
                # parse bam files
                #------------------------
                (rowwise_links, cov_profiles) = bamParser.parse(bamFiles,
                                                                con_names,
                                                                cid_2_indices)

                tot_cov = [sum(profile) for profile in cov_profiles]
                good_indices = np.nonzero(tot_cov)[0]
                bad_indices = [i for i in range(num_cons) if i not in good_indices]

                if len(bad_indices) > 0:
                    # report the bad contigs to the user
                    # and strip them before writing to the DB
                    print "****************************************************************"
                    print " IMPORTANT! - there are %d contigs with 0 coverage" % len(bad_indices)
                    print " across all stoits. They will be ignored:"
                    print "****************************************************************"
                    for i in xrange(0, min(5, len(bad_indices))):
                        print con_names[bad_indices[i]]
                    if len(bad_indices) > 5:
                      print '(+ %d additional contigs)' % (len(bad_indices)-5)
                    print "****************************************************************"

                    con_names = con_names[good_indices]
                    con_lengths = con_lengths[good_indices]
                    con_gcs = con_gcs[good_indices]
                    cov_profiles = cov_profiles[good_indices] 

                # these will need to be tupalized regardless...
                con_ksigs = [tuple(i) for i in con_ksigs[good_indices]]
                num_cons = len(con_names)

                #------------------------
                # calculate PCAs and write kmer sigs
                #------------------------
                # store the raw calculated kmer sigs in one table
                db_desc = []
                for mer in kse.kmerCols:
                     db_desc.append((mer, float))
                try:
                    h5file.createTable(profile_group,
                                       'kms',
                                       np.array(con_ksigs, dtype=db_desc),
                                       title='Kmer signatures',
                                       expectedrows=num_cons
                                       )
                except:
                    print "Error creating KMERSIG table:", exc_info()[0]
                    raise

                # compute the PCA of the ksigs and store these too
                pc_ksigs, sumvariance = conParser.PCAKSigs(con_ksigs)

                db_desc = []
                for i in xrange(0, len(pc_ksigs[0])):
                  db_desc.append(('pc' + str(i+1), float))

                try:
                    h5file.createTable(profile_group,
                                       'kpca',
                                       np.array(pc_ksigs, dtype=db_desc),
                                       title='Kmer signature PCAs',
                                       expectedrows=num_cons
                                       )
                except:
                    print "Error creating KMERVALS table:", exc_info()[0]
                    raise

                #------------------------
                # write cov profiles
                #------------------------
                # build a table template based on the number of bamfiles we have
                for i, bf in enumerate(bamFiles):
                    # assume the file is called something like "fred.bam"
                    # we want to rip off the ".bam" part
                    bam_desc = getBamDescriptor(bf, i + 1)
                    stoitColNames.append(bam_desc)

                stoitColNames = np_array(stoitColNames)
                
                # determine normalised cov profiles and shuffle the BAMs
                norm_coverages = np.array([np.linalg.norm(cov_profiles[i]) for i in range(num_cons)])
                CT = CoverageTransformer(num_cons,
                                         len(stoitColNames),
                                         norm_coverages,
                                         np.array(pc_ksigs)[:,0],
                                         cov_profiles,
                                         stoitColNames)
                
                CT.transformCP()
                # these will need to be tupalized regardless...
                cov_profiles = [tuple(i) for i in cov_profiles]
                CT.transformedCP = [tuple(i) for i in CT.transformedCP]
                CT.corners = [tuple(i) for i in CT.corners]
                # now CT stores the transformed coverages and other important information
                # the ordering of stoitColNames and cov_profiles should be fixed
                # so we will write this to the database without further modification

                # raw coverages
                db_desc = []
                for scn in CT.stoitColNames:
                    db_desc.append((scn, float)) 

                try:
                    h5file.createTable(profile_group,
                                       'coverage',
                                       np.array(cov_profiles, dtype=db_desc),
                                       title="Bam based coverage",
                                       expectedrows=num_cons)
                except:
                    print "Error creating coverage table:", exc_info()[0]
                    raise

                # transformed coverages
                db_desc = [('x', float),
                           ('y', float),
                           ('z', float)]
                try:
                    h5file.createTable(profile_group,
                                       'transCoverage',
                                       np.array(CT.transformedCP , dtype=db_desc),
                                       title="Transformed coverage",
                                       expectedrows=num_cons)
                except:
                    print "Error creating transformed coverage table:", exc_info()[0]
                    raise

                # transformed coverage corners
                db_desc = [('x', float),
                           ('y', float),
                           ('z', float)]
                try:
                    h5file.createTable(meta_group,
                                       'transCoverageCorners',
                                       np.array(CT.corners , dtype=db_desc),
                                       title="Transformed coverage corners",
                                       expectedrows=len(stoitColNames))
                except:
                    print "Error creating transformed coverage corner table:", exc_info()[0]
                    raise

                # normalised coverages
                db_desc = [('normCov', float)]
                try:
                    h5file.createTable(profile_group,
                                       'normCoverage',
                                       np.array(CT.normCoverages , dtype=db_desc),
                                       title="Normalised coverage",
                                       expectedrows=num_cons)
                except:
                    print "Error creating normalised coverage table:", exc_info()[0]
                    raise

                #------------------------
                # Add a table for the contigs
                #------------------------
                self.setBinAssignments((h5file, meta_group),
                                       image=zip(con_names,
                                                 [0]*num_cons,
                                                 con_lengths, con_gcs)
                                       )

                #------------------------
                # Add a table for the bins
                #------------------------
                self.setBinStats(dbFileName, [], firstWrite=True)

                print "    %s" % timer.getTimeStamp()

                #------------------------
                # contig links
                #------------------------
                # set table size according to the number of links returned from
                # the previous call
                db_desc = [('contig1', int),
                           ('contig2', int),
                           ('numReads', int),
                           ('linkType', int),
                           ('gap', int)]
                try:
                    h5file.createTable(links_group,
                                       'links',
                                       np.array(rowwise_links, dtype=db_desc),
                                       title="ContigLinks",
                                       expectedrows=len(rowwise_links))
                except:
                    print "Error creating links table:", exc_info()[0]
                    raise
                print "    %s" % timer.getTimeStamp()

                #------------------------
                # Add metadata
                #------------------------
                meta_data = (str.join(',',stoitColNames),
                             len(stoitColNames),
                             str.join(',',kse.kmerCols),
                             kmerSize,
                             len(kse.kmerCols),
                             num_cons,
                             0,
                             False,
                             False,
                             __current_GMDB_version__)
                self.setMeta(h5file, meta_data)

                # kmer signature variance table
                pc_var = [sumvariance[0]]
                for i in xrange(1, len(sumvariance)):
                  pc_var.append(sumvariance[i]-sumvariance[i-1])
                pc_var = tuple(pc_var)

                db_desc = []
                for i in xrange(0, len(pc_var)):
                    db_desc.append(('pc' + str(i+1) + '_var', float))

                try:
                    h5file.createTable(meta_group,
                                            'kpca_variance',
                                            np.array([pc_var], dtype=db_desc),
                                            title='Variance of kmer signature PCAs',
                                            expectedrows=1
                                            )
                except:
                    print "Error creating tmp_kpca_variance table:", exc_info()[0]
                    raise

        except:
            print "Error creating database:", dbFileName, exc_info()[0]
            raise

        print "****************************************************************"
        print "Data loaded successfully!"
        print " ->",num_cons,"contigs"
        print " ->",len(stoitColNames),"BAM files"
        print "Written to: '"+dbFileName+"'"
        print "****************************************************************"
        print "    %s" % timer.getTimeStamp()

        # all good!
        return True

    def promptOnOverwrite(self, dbFileName, minimal=False):
        """Check that the user is ok with overwriting the db"""
        input_not_ok = True
        valid_responses = ['Y','N']
        vrs = ",".join([str.lower(str(x)) for x in valid_responses])
        while(input_not_ok):
            if(minimal):
                option = raw_input(" Overwrite? ("+vrs+") : ")
            else:

                option = raw_input(" ****WARNING**** Database: '"+dbFileName+"' exists.\n" \
                                   " If you continue you *WILL* delete any previous analyses!\n" \
                                   " Overwrite? ("+vrs+") : ")
            if(option.upper() in valid_responses):
                print "****************************************************************"
                return option.upper()
            else:
                print "Error, unrecognised choice '"+option.upper()+"'"
                minimal = True

#------------------------------------------------------------------------------
# DB UPGRADE

    def checkAndUpgradeDB(self, dbFileName, silent=False):
        """Check the DB and upgrade if necessary"""
        # get the DB format version
        this_DB_version = self.getGMDBFormat(dbFileName)
        if __current_GMDB_version__ == this_DB_version:
            if not silent:
                print "    GroopM DB version (%s) up to date" % this_DB_version
            return

        # now, if we get here then we need to do some work
        upgrade_tasks = {}
        upgrade_tasks[(0,1)] = self.upgradeDB_0_to_1
        upgrade_tasks[(1,2)] = self.upgradeDB_1_to_2
        upgrade_tasks[(2,3)] = self.upgradeDB_2_to_3
        upgrade_tasks[(3,4)] = self.upgradeDB_3_to_4
        upgrade_tasks[(4,5)] = self.upgradeDB_4_to_5

        # we need to apply upgrades in order!
        # keep applying the upgrades as long as we need to
        while this_DB_version < __current_GMDB_version__:
            task = (this_DB_version, this_DB_version+1)
            upgrade_tasks[task](dbFileName)
            this_DB_version += 1

    def upgradeDB_0_to_1(self, dbFileName):
        """Upgrade a GM db from version 0 to version 1"""
        print "*******************************************************************************\n"
        print "              *** Upgrading GM DB from version 0 to version 1 ***"
        print ""
        print "                            please be patient..."
        print ""
        # the change in this version is that we'll be saving the first
        # two kmerSig PCA's in a separate table
        print "    Calculating and storing the kmerSig PCAs"

        # compute the PCA of the ksigs
        ksigs = self.getKmerSigs(dbFileName)
        CP = ContigParser()
        pc_ksigs, sumvariance = CP.PCAKSigs(ksigs)
        num_cons = len(pc_ksigs)

        db_desc = [('pc1', float),
                   ('pc2', float)]
        try:
            with tables.openFile(dbFileName, mode='a', rootUEP="/profile") as profile_group:
                try:
                    profile_group.createTable('/',
                                              'kpca',
                                              np.array(pc_ksigs, dtype=db_desc),
                                              title='Kmer signature PCAs',
                                              expectedrows=num_cons
                                              )
                except:
                    print "Error creating KMERVALS table:", exc_info()[0]
                    raise
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

        # update the formatVersion field and we're done
        self.setGMDBFormat(dbFileName, 1)
        print "*******************************************************************************"

    def upgradeDB_1_to_2(self, dbFileName):
        """Upgrade a GM db from version 1 to version 2"""
        print "*******************************************************************************\n"
        print "              *** Upgrading GM DB from version 1 to version 2 ***"
        print ""
        print "                            please be patient..."
        print ""
        # the change in this version is that we'll be saving a variable number of kmerSig PCA's
        # and GC information for each contig
        print "    Calculating and storing the kmer signature PCAs"

        # grab any data needed from database before opening if for modification
        bin_ids = self.getBins(dbFileName)
        orig_con_names = self.getContigNames(dbFileName)
        ksigs = self.getKmerSigs(dbFileName)

        # compute the PCA of the ksigs
        conParser = ContigParser()
        pc_ksigs, sumvariance = conParser.PCAKSigs(ksigs)
        num_cons = len(pc_ksigs)

        db_desc = []
        for i in xrange(0, len(pc_ksigs[0])):
          db_desc.append(('pc' + str(i+1), float))

        try:
            with tables.openFile(dbFileName, mode='a', rootUEP="/") as h5file:
                pg = h5file.getNode('/', name='profile')
                try:
                    try:
                        h5file.removeNode(pg, 'tmp_kpca')
                    except:
                        pass

                    h5file.createTable(pg,
                                       'tmp_kpca',
                                       np.array(pc_ksigs, dtype=db_desc),
                                       title='Kmer signature PCAs',
                                       expectedrows=num_cons
                                      )

                    h5file.renameNode(pg, 'kpca', 'tmp_kpca', overwrite=True)

                except:
                    print "Error creating kpca table:", exc_info()[0]
                    raise

                # Add GC
                contigFile = raw_input('\nPlease specify fasta file containing the bam reference sequences: ')
                with open(contigFile, "r") as f:
                    try:
                        contigInfo = {}
                        for cid,seq,qual in conParser.readfq(f):
                            contigInfo[cid] = (len(seq), conParser.calculateGC(seq))

                        # sort the contig names here once!
                        con_names = np.array(sorted(contigInfo.keys()))

                        # keep everything in order...
                        con_gcs = np.array([contigInfo[cid][1] for cid in con_names])
                        con_lengths = np.array([contigInfo[cid][0] for cid in con_names])
                    except:
                        print "Error parsing contigs"
                        raise

                # remove any contigs not in the current DB (these were removed due to having zero coverage)
                good_indices = [i for i in range(len(orig_con_names)) if orig_con_names[i] in con_names]

                con_names = con_names[good_indices]
                con_lengths = con_lengths[good_indices]
                con_gcs = con_gcs[good_indices]
                bin_ids = bin_ids[good_indices]

                mg = h5file.getNode('/', name='meta')
                self.setBinAssignments((h5file, mg),
                               image=zip(con_names,
                                         bin_ids,
                                         con_lengths, con_gcs)
                               )
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

        # update the formatVersion field and we're done
        self.setGMDBFormat(dbFileName, 2)
        print "*******************************************************************************"

    def upgradeDB_2_to_3(self, dbFileName):
        """Upgrade a GM db from version 2 to version 3"""
        print "*******************************************************************************\n"
        print "              *** Upgrading GM DB from version 2 to version 3 ***"
        print ""
        print "                            please be patient..."
        print ""
        # the change in this version is that we'll be saving the variance for each kmerSig PCA
        print "    Calculating and storing variance of kmer signature PCAs"

        # compute the PCA of the ksigs
        conParser = ContigParser()
        ksigs = self.getKmerSigs(dbFileName)
        pc_ksigs, sumvariance = conParser.PCAKSigs(ksigs)

        # calcualte variance of each PC
        pc_var = [sumvariance[0]]
        for i in xrange(1, len(sumvariance)):
          pc_var.append(sumvariance[i]-sumvariance[i-1])
        pc_var = tuple(pc_var)

        db_desc = []
        for i in xrange(0, len(pc_var)):
          db_desc.append(('pc' + str(i+1) + '_var', float))

        try:
            with tables.openFile(dbFileName, mode='a', rootUEP="/") as h5file:
                meta = h5file.getNode('/', name='meta')
                try:
                    try:
                        h5file.removeNode(meta, 'tmp_kpca_variance')
                    except:
                        pass

                    h5file.createTable(meta,
                                              'tmp_kpca_variance',
                                              np.array([pc_var], dtype=db_desc),
                                              title='Variance of kmer signature PCAs',
                                              expectedrows=1
                                              )

                    h5file.renameNode(meta, 'kpca_variance', 'tmp_kpca_variance', overwrite=True)

                except:
                    print "Error creating kpca_variance table:", exc_info()[0]
                    raise
        except:
            print "Error opening DB:", dbFileName, exc_info()[0]
            raise

        # update the formatVersion field and we're done
        self.setGMDBFormat(dbFileName, 3)
        print "*******************************************************************************"

    def upgradeDB_3_to_4(self, dbFileName):
        """Upgrade a GM db from version 3 to version 4"""
        print "*******************************************************************************\n"
        print "              *** Upgrading GM DB from version 3 to version 4 ***"
        print ""
        print "                            please be patient..."
        print ""
        # the change in this version is that we'll be saving the variance for each kmerSig PCA
        print "    Adding chimeric flag for each bin."
        print "    !!! Groopm core must be run again for this flag to be properly set. !!!"

        # read existing data in 'bins' table
        try:
            with tables.openFile(dbFileName, mode='r') as h5file:
                ret_dict = {}
                all_rows = h5file.root.meta.bins.read()
                for row in all_rows:
                    ret_dict[row[0]] = row[1]
        except:
            print "Error opening DB:", dbFileName, exc_info()[0]
            raise

        # write new table with chimeric flag set to False by default
        db_desc = [('bid', int),
                   ('numMembers', int),
                   ('isLikelyChimeric', bool)]

        data = []
        for bid in ret_dict:
          data.append((bid, ret_dict[bid], False))

        bd = np.array(data, dtype=db_desc)

        try:
            with tables.openFile(dbFileName, mode='a', rootUEP="/") as h5file:
                mg = h5file.getNode('/', name='meta')

                try:
                    h5file.removeNode(mg, 'tmp_bins')
                except:
                    pass

                try:
                    h5file.createTable(mg,
                                       'tmp_bins',
                                       bd,
                                       title="Bin information",
                                       expectedrows=1)
                except:
                    print "Error creating META table:", exc_info()[0]
                    raise

                h5file.renameNode(mg, 'bins', 'tmp_bins', overwrite=True)
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

        # update the formatVersion field and we're done
        self.setGMDBFormat(dbFileName, 4)
        print "*******************************************************************************"

    def upgradeDB_4_to_5(self, dbFileName):
        """Upgrade a GM db from version 4 to version 5"""
        print "*******************************************************************************\n"
        print "              *** Upgrading GM DB from version 4 to version 5 ***"
        print ""
        print "                            please be patient..."
        print ""
        # the change in this version is that we'll be saving the transformed coverage coords
        print "    Saving transformed coverage profiles"
        print "    You will not need to re-run parse or core due to this change"

        # we need to get the raw coverage profiles and the kmerPCA1 data
        indices = self.getConditionalIndices(dbFileName, silent=False, checkUpgrade=False)
        kPCA_1 = self.getKmerPCAs(dbFileName, indices=indices)[:,0]
        raw_coverages = self.getCoverageProfiles(dbFileName, indices=indices)
        norm_coverages = np.array([np.linalg.norm(raw_coverages[i]) for i in range(len(indices))])
        
        CT = CoverageTransformer(len(indices),
                                 self.getNumStoits(dbFileName),
                                 norm_coverages,
                                 kPCA_1,
                                 raw_coverages,
                                 np.array(self.getStoitColNames(dbFileName).split(",")))
        
        CT.transformCP()
        CT.transformedCP = [tuple(i) for i in CT.transformedCP]
        CT.covProfiles = [tuple(i) for i in CT.covProfiles]
        CT.corners = [tuple(i) for i in CT.corners]
        
        # now CT stores the transformed coverages and other important information
        # we will write this to the database
        with tables.openFile(dbFileName, mode='a', rootUEP="/") as h5file:
            meta_group = h5file.getNode('/', name='meta')
            profile_group = h5file.getNode('/', name='profile')

            # raw coverages - we may have reordered rows, so we should fix this now!
            db_desc = []
            for scn in CT.stoitColNames:
                db_desc.append((scn, float)) 
            
            try:
                h5file.removeNode(mg, 'tmp_coverages')
            except:
                pass

            try:
                h5file.createTable(profile_group,
                                   'tmp_coverages',
                                   np.array(CT.covProfiles, dtype=db_desc),
                                   title="Bam based coverage",
                                   expectedrows=CT.numContigs)
            except:
                print "Error creating coverage table:", exc_info()[0]
                raise

            h5file.renameNode(profile_group, 'coverage', 'tmp_coverages', overwrite=True)
        
            # transformed coverages
            db_desc = [('x', float),
                       ('y', float),
                       ('z', float)]
            try:
                h5file.createTable(profile_group,
                                   'transCoverage',
                                   np.array(CT.transformedCP , dtype=db_desc),
                                   title="Transformed coverage",
                                   expectedrows=CT.numContigs)
            except:
                print "Error creating transformed coverage table:", exc_info()[0]
                raise
    
            # transformed coverage corners
            db_desc = [('x', float),
                       ('y', float),
                       ('z', float)]
            try:
                h5file.createTable(meta_group,
                                   'transCoverageCorners',
                                   np.array(CT.corners , dtype=db_desc),
                                   title="Transformed coverage corners",
                                   expectedrows=CT.numStoits)
            except:
                print "Error creating transformed coverage corner table:", exc_info()[0]
                raise
    
    
            # normalised coverages
            db_desc = [('normCov', float)]
            try:
                h5file.createTable(profile_group,
                                   'normCoverage',
                                   np.array(CT.normCoverages , dtype=db_desc),
                                   title="Normalised coverage",
                                   expectedrows=CT.numContigs)
            except:
                print "Error creating normalised coverage table:", exc_info()[0]
                raise

        # stoit col names may have been shuffled
        meta_data = (",".join([str(i) for i in CT.stoitColNames]),
                    CT.numStoits,
                    self.getMerColNames(dbFileName),
                    self.getMerSize(dbFileName),
                    self.getNumMers(dbFileName),
                    self.getNumCons(dbFileName),
                    self.getNumBins(dbFileName),
                    self.isClustered(dbFileName),
                    self.isComplete(dbFileName),
                    self.getGMDBFormat(dbFileName))

        with tables.openFile(dbFileName, mode='a', rootUEP="/") as h5file:
            self.setMeta(h5file, meta_data, overwrite=True)
        
        # update the formatVersion field and we're done
        self.setGMDBFormat(dbFileName, 5)
        print "*******************************************************************************"


#------------------------------------------------------------------------------
# GET LINKS

    def restoreLinks(self, dbFileName, indices=[], silent=False):
        """Restore the links hash for a given set of indices"""
        full_record = []
        try:
            with tables.openFile(dbFileName, mode='r') as h5file:
                full_record = [list(x) for x in h5file.root.links.links.readWhere("contig1 >= 0")]
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

        if indices == []:
            # get all!
            indices = self.getConditionalIndices(dbFileName, silent=silent)

        links_hash = {}
        if full_record != []:
            for record in full_record:
                # make sure we have storage
                if record[0] in indices and record[1] in indices:
                    try:
                        links_hash[record[0]].append(record[1:])
                    except KeyError:
                        links_hash[record[0]] = [record[1:]]
        return links_hash

#------------------------------------------------------------------------------
# GET / SET DATA TABLES - PROFILES

    def getConditionalIndices(self, dbFileName, condition='', silent=False, checkUpgrade=True):
        """return the indices into the db which meet the condition"""
        # check the DB out and see if we need to change anything about it
        if checkUpgrade:
            self.checkAndUpgradeDB(dbFileName, silent=silent)

        if('' == condition):
            condition = "cid != ''" # no condition breaks everything!
        try:
            with tables.openFile(dbFileName, mode='r') as h5file:
                return np.array([x.nrow for x in h5file.root.meta.contigs.where(condition)])
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

    def getCoverageProfiles(self, dbFileName, condition='', indices=np.array([])):
        """Load coverage profiles"""
        try:
            with tables.openFile(dbFileName, mode='r') as h5file:
                if(np.size(indices) != 0):
                    return np.array([list(h5file.root.profile.coverage[x]) for x in indices])
                else:
                    if('' == condition):
                        condition = "cid != ''" # no condition breaks everything!
                    return np.array([list(h5file.root.profile.coverage[x.nrow]) for x in h5file.root.meta.contigs.where(condition)])
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

    def getTransformedCoverageProfiles(self, dbFileName, condition='', indices=np.array([])):
        """Load transformed coverage profiles"""
        try:
            with tables.openFile(dbFileName, mode='r') as h5file:
                if(np.size(indices) != 0):
                    return np.array([list(h5file.root.profile.transCoverage[x]) for x in indices])
                else:
                    if('' == condition):
                        condition = "cid != ''" # no condition breaks everything!
                    return np.array([list(h5file.root.profile.transCoverage[x.nrow]) for x in h5file.root.meta.contigs.where(condition)])
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

    def getNormalisedCoverageProfiles(self, dbFileName, condition='', indices=np.array([])):
        """Load normalised coverage profiles"""
        try:
            with tables.openFile(dbFileName, mode='r') as h5file:
                if(np.size(indices) != 0):
                    return np.array([list(h5file.root.profile.normCoverage[x]) for x in indices])
                else:
                    if('' == condition):
                        condition = "cid != ''" # no condition breaks everything!
                    return np.array([list(h5file.root.profile.normCoverage[x.nrow]) for x in h5file.root.meta.contigs.where(condition)])
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

    def nukeBins(self, dbFileName):
        """Reset all bin information, completely"""
        print "    Clearing all old bin information from",dbFileName
        self.setBinStats(dbFileName, [])
        self.setNumBins(dbFileName, 0)
        self.setBinAssignments(dbFileName, updates={}, nuke=True)

    def setBinStats(self, dbFileName, updates, firstWrite=False):
        """Set bins table

        updates is a list of tuples which looks like:
        [ (bid, numMembers, isLikelyChimeric) ]
        """

        db_desc = [('bid', int),
                   ('numMembers', int),
                   ('isLikelyChimeric', bool)]
        bd = np.array(updates, dtype=db_desc)

        try:
            with tables.openFile(dbFileName, mode='a', rootUEP="/") as h5file:
                mg = h5file.getNode('/', name='meta')
                if not firstWrite:
                    t_name = 'tmp_bins'
                    # nuke any previous failed attempts
                    try:
                        h5file.removeNode(mg, 'tmp_bins')
                    except:
                        pass
                else:
                    t_name = 'bins'

                try:
                    h5file.createTable(mg,
                                       t_name,
                                       bd,
                                       title="Bin information",
                                       expectedrows=1)
                except:
                    print "Error creating META table:", exc_info()[0]
                    raise

                if not firstWrite:
                    # rename the tmp table to overwrite
                    h5file.renameNode(mg, 'bins', 'tmp_bins', overwrite=True)
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

    def getBinStats(self, dbFileName):
        """Load data from bins table

        Returns a dict of type:
        { bid : [numMembers, isLikelyChimeric] }
        """
        try:
            with tables.openFile(dbFileName, mode='r') as h5file:
                ret_dict = {}
                all_rows = h5file.root.meta.bins.read()
                for row in all_rows:
                    ret_dict[row[0]] = [row[1], row[2]]

                return ret_dict
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise
        return {}

    def getBins(self, dbFileName, condition='', indices=np.array([])):
        """Load per-contig bins"""
        try:
            with tables.openFile(dbFileName, mode='r') as h5file:
                if(np.size(indices) != 0):
                    return np.array([h5file.root.meta.contigs[x][1] for x in indices]).ravel()
                else:
                    if('' == condition):
                        condition = "cid != ''" # no condition breaks everything!
                    return np.array([list(x)[1] for x in h5file.root.meta.contigs.readWhere(condition)]).ravel()
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

    def setBinAssignments(self, storage, updates=None, image=None, nuke=False):
        """Set per-contig bins

        updates is a dictionary which looks like:
        { tableRow : binValue }
        if updates is set then storage is the
        path to the hdf file

        image is a list of tuples which look like:
        [(cid, bid, len, gc)]
        if image is set then storage is a tuple of type:
        (h5file, group)
        """
        db_desc = [('cid', '|S512'),
                   ('bid', int),
                   ('length', int),
                   ('gc', float)]
        closeh5 = False
        if updates is not None:
            # we need to build the image
            dbFileName = storage
            contig_names = self.getContigNames(dbFileName)
            contig_lengths = self.getContigLengths(dbFileName)
            contig_gcs = self.getContigGCs(dbFileName)
            num_cons = len(contig_lengths)
            if nuke:
                # clear all bin assignments
                bins = [0]*num_cons
            else:
                bins = self.getBins(dbFileName)

            # now apply the updates
            for tr in updates.keys():
                bins[tr] = updates[tr]

            # and build the image
            image = np.array(zip(contig_names, bins, contig_lengths, contig_gcs),
                             dtype=db_desc)

            try:
                h5file = tables.openFile(dbFileName, mode='a')
            except:
                print "Error opening DB:",dbFileName, exc_info()[0]
                raise
            meta_group = h5file.getNode('/', name='meta')
            closeh5 = True

        elif image is not None:
            h5file = storage[0]
            meta_group = storage[1]
            num_cons = len(image)
            image = np.array(image,
                             dtype=db_desc)
        else:
            print "get with the program dude"
            return

        # now we write the data
        try:
            # get rid of any failed attempts
            h5file.removeNode(meta_group, 'tmp_contigs')
        except:
            pass

        try:
            h5file.createTable(meta_group,
                               'tmp_contigs',
                               image,
                               title="Contig information",
                               expectedrows=num_cons)
        except:
            print "Error creating CONTIG table:", exc_info()[0]
            raise

        # rename the tmp table to overwrite
        h5file.renameNode(meta_group, 'contigs', 'tmp_contigs', overwrite=True)
        if closeh5:
            h5file.close()

    def getContigNames(self, dbFileName, condition='', indices=np.array([])):
        """Load contig names"""
        try:
            with tables.openFile(dbFileName, mode='r') as h5file:
                if(np.size(indices) != 0):
                    return np.array([h5file.root.meta.contigs[x][0] for x in indices]).ravel()
                else:
                    if('' == condition):
                        condition = "cid != ''" # no condition breaks everything!
                    return np.array([list(x)[0] for x in h5file.root.meta.contigs.readWhere(condition)]).ravel()
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

    def getContigLengths(self, dbFileName, condition='', indices=np.array([])):
        """Load contig lengths"""
        try:
            with tables.openFile(dbFileName, mode='r') as h5file:
                if(np.size(indices) != 0):
                    return np.array([h5file.root.meta.contigs[x][2] for x in indices]).ravel()
                else:
                    if('' == condition):
                        condition = "cid != ''" # no condition breaks everything!
                    return np.array([list(x)[2] for x in h5file.root.meta.contigs.readWhere(condition)]).ravel()
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

    def getContigGCs(self, dbFileName, condition='', indices=np.array([])):
        """Load contig gcs"""
        try:
            with tables.openFile(dbFileName, mode='r') as h5file:
                if(np.size(indices) != 0):
                    return np.array([h5file.root.meta.contigs[x][3] for x in indices]).ravel()
                else:
                    if('' == condition):
                        condition = "cid != ''" # no condition breaks everything!
                    return np.array([list(x)[3] for x in h5file.root.meta.contigs.readWhere(condition)]).ravel()
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

    def getKmerSigs(self, dbFileName, condition='', indices=np.array([])):
        """Load kmer sigs"""
        try:
            with tables.openFile(dbFileName, mode='r') as h5file:
                if(np.size(indices) != 0):
                    return np.array([list(h5file.root.profile.kms[x]) for x in indices])
                else:
                    if('' == condition):
                        condition = "cid != ''" # no condition breaks everything!
                    return np.array([list(h5file.root.profile.kms[x.nrow]) for x in h5file.root.meta.contigs.where(condition)])
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

    def getKmerPCAs(self, dbFileName, condition='', indices=np.array([])):
        """Load kmer sig PCAs"""
        try:
            with tables.openFile(dbFileName, mode='r') as h5file:
                if(np.size(indices) != 0):
                    return np.array([list(h5file.root.profile.kpca[x]) for x in indices])
                else:
                    if('' == condition):
                        condition = "cid != ''" # no condition breaks everything!
                    return np.array([list(h5file.root.profile.kpca[x.nrow]) for x in h5file.root.meta.contigs.where(condition)])
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

#------------------------------------------------------------------------------
# GET / SET METADATA

    def getKmerVarPC(self, dbFileName, condition='', indices=np.array([])):
        """Load variance of kmer sig PCAs"""
        try:
            with tables.openFile(dbFileName, mode='r') as h5file:
              return np.array(h5file.root.meta.kpca_variance)[0]
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

    def getTransformedCoverageCorners(self, dbFileName):
        """Load transformed coverage corners"""
        try:
            with tables.openFile(dbFileName, mode='r') as h5file:
                return np.array([list(x) for x in h5file.root.meta.transCoverageCorners.read()])
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

    def setMeta(self, h5file, metaData, overwrite=False):
        """Write metadata into the table

        metaData should be a tuple of values
        """
        db_desc = [('stoitColNames', '|S512'),
                   ('numStoits', int),
                   ('merColNames', '|S4096'),
                   ('merSize', int),
                   ('numMers', int),
                   ('numCons', int),
                   ('numBins', int),
                   ('clustered', bool),     # set to true after clustering is complete
                   ('complete', bool),      # set to true after clustering finishing is complete
                   ('formatVersion', int)]
        md = np.array([metaData], dtype=db_desc)

        # get hold of the group
        mg = h5file.getNode('/', name='meta')

        if overwrite:
            t_name = 'tmp_meta'
            # nuke any previous failed attempts
            try:
                h5file.removeNode(mg, 'tmp_meta')
            except:
                pass
        else:
            t_name = 'meta'

        try:
            h5file.createTable(mg,
                               t_name,
                               md,
                               "Descriptive data",
                               expectedrows=1)
        except:
            print "Error creating META table:", exc_info()[0]
            raise

        if overwrite:
            # rename the tmp table to overwrite
            h5file.renameNode(mg, 'meta', 'tmp_meta', overwrite=True)

    def getMetaField(self, dbFileName, fieldName):
        """return the value of fieldName in the metadata tables"""
        try:
            with tables.openFile(dbFileName, mode='r') as h5file:
                # theres only one value
                return h5file.root.meta.meta.read()[fieldName][0]
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

    def setGMDBFormat(self, dbFileName, version):
        """Update the GMDB format version"""
        stoit_col_names = self.getStoitColNames(dbFileName)
        meta_data = (stoit_col_names,
                    len(stoit_col_names.split(',')),
                    self.getMerColNames(dbFileName),
                    self.getMerSize(dbFileName),
                    self.getNumMers(dbFileName),
                    self.getNumCons(dbFileName),
                    self.getNumBins(dbFileName),
                    self.isClustered(dbFileName),
                    self.isComplete(dbFileName),
                    version)
        try:
            with tables.openFile(dbFileName, mode='a', rootUEP="/") as h5file:
                self.setMeta(h5file, meta_data, overwrite=True)
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

    def getGMDBFormat(self, dbFileName):
        """return the format version of this GM file"""
        # this guy needs to be a bit different to the other meta methods
        # becuase earlier versions of GM didn't include a format parameter
        with tables.openFile(dbFileName, mode='r') as h5file:
            # theres only one value
            try:
                this_DB_version = h5file.root.meta.meta.read()['formatVersion'][0]
            except ValueError:
                # this happens when an oldskool formatless DB is loaded
                this_DB_version = 0
        return this_DB_version

    def getNumStoits(self, dbFileName):
        """return the value of numStoits in the metadata tables"""
        return self.getMetaField(dbFileName, 'numStoits')

    def getMerColNames(self, dbFileName):
        """return the value of merColNames in the metadata tables"""
        return self.getMetaField(dbFileName, 'merColNames')

    def getMerSize(self, dbFileName):
        """return the value of merSize in the metadata tables"""
        return self.getMetaField(dbFileName, 'merSize')

    def getNumMers(self, dbFileName):
        """return the value of numMers in the metadata tables"""
        return self.getMetaField(dbFileName, 'numMers')

    def getNumCons(self, dbFileName):
        """return the value of numCons in the metadata tables"""
        return self.getMetaField(dbFileName, 'numCons')

    def setNumBins(self, dbFileName, numBins):
        """set the number of bins"""
        stoit_col_names = self.getStoitColNames(dbFileName)
        meta_data = (stoit_col_names,
                    len(stoit_col_names.split(',')),
                    self.getMerColNames(dbFileName),
                    self.getMerSize(dbFileName),
                    self.getNumMers(dbFileName),
                    self.getNumCons(dbFileName),
                    numBins,
                    self.isClustered(dbFileName),
                    self.isComplete(dbFileName),
                    self.getGMDBFormat(dbFileName))
        try:
            with tables.openFile(dbFileName, mode='a', rootUEP="/") as h5file:
                self.setMeta(h5file, meta_data, overwrite=True)
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

    def getNumBins(self, dbFileName):
        """return the value of numBins in the metadata tables"""
        return self.getMetaField(dbFileName, 'numBins')

    def getStoitColNames(self, dbFileName):
        """return the value of stoitColNames in the metadata tables"""
        return self.getMetaField(dbFileName, 'stoitColNames')

#------------------------------------------------------------------------------
# GET / SET WORKFLOW FLAGS

    def isClustered(self, dbFileName):
        """Has this data set been clustered?"""
        try:
            with tables.openFile(dbFileName, mode='r') as h5file:
                return h5file.root.meta.meta.read()['clustered']
        except:
            print "Error opening database:", dbFileName, exc_info()[0]
            raise

    def setClustered(self, dbFileName, state):
        """Set the state of clustering"""
        stoit_col_names = self.getStoitColNames(dbFileName)
        meta_data = (stoit_col_names,
                    len(stoit_col_names.split(',')),
                    self.getMerColNames(dbFileName),
                    self.getMerSize(dbFileName),
                    self.getNumMers(dbFileName),
                    self.getNumCons(dbFileName),
                    self.getNumBins(dbFileName),
                    state,
                    self.isComplete(dbFileName),
                    self.getGMDBFormat(dbFileName))
        try:
            with tables.openFile(dbFileName, mode='a', rootUEP="/") as h5file:
                self.setMeta(h5file, meta_data, overwrite=True)
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

    def isComplete(self, dbFileName):
        """Has this data set been *completely* clustered?"""
        try:
            with tables.openFile(dbFileName, mode='r') as h5file:
                return h5file.root.meta.meta.read()['complete']
        except:
            print "Error opening database:", dbFileName, exc_info()[0]
            raise

    def setComplete(self, dbFileName, state):
        """Set the state of completion"""
        stoit_col_names = self.getStoitColNames(dbFileName)
        meta_data = (stoit_col_names,
                    len(stoit_col_names.split(',')),
                    self.getMerColNames(dbFileName),
                    self.getMerSize(dbFileName),
                    self.getNumMers(dbFileName),
                    self.getNumCons(dbFileName),
                    self.getNumBins(dbFileName),
                    self.isClustered(dbFileName),
                    state,
                    self.getGMDBFormat(dbFileName))
        try:
            with tables.openFile(dbFileName, mode='a', rootUEP="/") as h5file:
                self.setMeta(h5file, meta_data, overwrite=True)
        except:
            print "Error opening DB:",dbFileName, exc_info()[0]
            raise

#------------------------------------------------------------------------------
# FILE / IO

    def dumpData(self, dbFileName, fields, outFile, separator, useHeaders):
        """Dump data to file"""
        header_strings = []
        data_arrays = []

        if fields == ['all']:
            fields = ['names', 'lengths', 'gc', 'bins', 'coverage', 'tcoverage', 'ncoverage', 'mers']

        num_fields = len(fields)
        data_converters = []

        for field in fields:
            if field == 'names':
                header_strings.append('cid')
                data_arrays.append(self.getContigNames(dbFileName))
                data_converters.append(lambda x : x)

            elif field == 'lengths':
                header_strings.append('length')
                data_arrays.append(self.getContigLengths(dbFileName))
                data_converters.append(lambda x : str(x))

            elif field == 'gc':
                header_strings.append('GCs')
                data_arrays.append(self.getContigGCs(dbFileName))
                data_converters.append(lambda x : str(x))

            elif field == 'bins':
                header_strings.append('bid')
                data_arrays.append(self.getBins(dbFileName))
                data_converters.append(lambda x : str(x))

            elif field == 'coverage':
                stoits = self.getStoitColNames(dbFileName).split(',')
                for stoit in stoits:
                    header_strings.append(stoit)
                data_arrays.append(self.getCoverageProfiles(dbFileName))
                data_converters.append(lambda x : separator.join(["%0.4f" % i for i in x]))

            elif field == 'tcoverage':
                header_strings.append('transformedCoverageX')
                header_strings.append('transformedCoverageY')
                header_strings.append('transformedCoverageZ')
                data_arrays.append(self.getTransformedCoverageProfiles(dbFileName))
                data_converters.append(lambda x : separator.join(["%0.4f" % i for i in x]))

            elif field == 'ncoverage':
                header_strings.append('normalisedCoverage')
                data_arrays.append(self.getNormalisedCoverageProfiles(dbFileName))
                data_converters.append(lambda x : separator.join(["%0.4f" % i for i in x]))

            elif field == 'mers':
                mers = self.getMerColNames(dbFileName).split(',')
                for mer in mers:
                    header_strings.append(mer)
                data_arrays.append(self.getKmerSigs(dbFileName))
                data_converters.append(lambda x : separator.join(["%0.4f" % i for i in x]))

        try:
            with open(outFile, 'w') as fh:
                if useHeaders:
                    header = separator.join(header_strings) + "\n"
                    fh.write(header)

                num_rows = len(data_arrays[0])
                for i in range(num_rows):
                    fh.write(data_converters[0](data_arrays[0][i]))
                    for j in range(1, num_fields):
                        fh.write(separator+data_converters[j](data_arrays[j][i]))
                    fh.write('\n')
        except:
            print "Error opening output file %s for writing" % outFile
            raise

###############################################################################
###############################################################################
###############################################################################
###############################################################################
class ContigParser:
    """Main class for reading in and parsing contigs"""
    def __init__(self): pass

    def readfq(self, fp): # this is a generator function
        """https://github.com/lh3"""
        last = None # this is a buffer keeping the last unprocessed line
        while True: # mimic closure; is it a bad idea?
            if not last: # the first record or a record following a fastq
                for l in fp: # search for the start of the next record
                    if l[0] in '>@': # fasta/q header line
                        last = l[:-1] # save this line
                        break
            if not last: break
            name, seqs, last = last[1:].split()[0], [], None
            for l in fp: # read the sequence
                if l[0] in '@+>':
                    last = l[:-1]
                    break
                seqs.append(l[:-1])
            if not last or last[0] != '+': # this is a fasta record
                yield name, ''.join(seqs), None # yield a fasta record
                if not last: break
            else: # this is a fastq record
                seq, leng, seqs = ''.join(seqs), 0, []
                for l in fp: # read the quality
                    seqs.append(l[:-1])
                    leng += len(l) - 1
                    if leng >= len(seq): # have read enough quality
                        last = None
                        yield name, seq, ''.join(seqs); # yield a fastq record
                        break
                if last: # reach EOF before reading enough quality
                    yield name, seq, None # yield a fasta record instead
                    break

    def parse(self, contigFile, kse):
        """Do the heavy lifting of parsing"""
        print "Parsing contigs"
        contigInfo = {} # save everything here first so we can sort accordingly
        for cid,seq,qual in self.readfq(contigFile):
            contigInfo[cid] = (kse.getKSig(seq.upper()), len(seq), self.calculateGC(seq))

        # sort the contig names here once!
        con_names = np.array(sorted(contigInfo.keys()))

        # keep everything in order...
        con_gcs = np.array([contigInfo[cid][2] for cid in con_names])
        con_lengths = np.array([contigInfo[cid][1] for cid in con_names])
        con_ksigs = np.array([contigInfo[cid][0] for cid in con_names])

        return (con_names, con_gcs, con_lengths, con_ksigs)

        # store the PCA'd kmersigs
        k_PCA_data = np.reshape(k_PCA_data, (rows,cols))
        self.storeSigPCAs(k_PCA_data, kPCATable)

    def calculateGC(self, seq):
      """Calculate fraction of nucleotides that are G or C."""
      testSeq = seq.upper()
      gc = testSeq.count('G') + testSeq.count('C')
      at = testSeq.count('A') + testSeq.count('T')

      return float(gc) / (gc + at)

    def PCAKSigs(self, kSigs, variance = 0.8):
        """PCA kmer sig data. All PCs require to capture the specified variance are returned.

        returns an array of tuples [(pc11, pc21, ..., pcN1), (pc12, pc22, ..., pnN2), ...]
        """

        # make a copy
        data = np.copy(kSigs)

        Center(data,verbose=0)
        p = PCA(data, fraction=variance)
        components = p.pc()

        return [tuple(i) for i in components], p.sumvariance[0:len(components[0])]

    def getWantedSeqs(self, contigFile, wanted, storage={}):
        """Do the heavy lifting of parsing"""
        print "Parsing contigs"
        for cid,seq,qual in self.readfq(contigFile):
            if(cid in wanted):
                storage[cid] = seq
        return storage

###############################################################################
###############################################################################
###############################################################################
###############################################################################
class KmerSigEngine:
    """Simple class for determining kmer signatures"""
    def __init__(self, kLen=4):
        self.kLen = kLen
        self.compl = s_maketrans('ACGT', 'TGCA')
        (self.kmerCols, self.llDict) = self.makeKmerColNames(makeLL=True)
        self.numMers = len(self.kmerCols)

    def makeKmerColNames(self, makeLL=False):
        """Work out the range of kmers required based on kmer length

        returns a list of sorted kmers and optionally a llo dict
        """
        # build up the big list
        base_words = ("A","C","G","T")
        out_list = ["A","C","G","T"]
        for i in range(1,self.kLen):
            working_list = []
            for mer in out_list:
                for char in base_words:
                    working_list.append(mer+char)
            out_list = working_list

        # pare it down based on lexicographical ordering
        ret_list = []
        ll_dict = {}
        for mer in out_list:
            lmer = self.shiftLowLexi(mer)
            ll_dict[mer] = lmer
            if lmer not in ret_list:
                ret_list.append(lmer)
        if makeLL:
            return (sorted(ret_list), ll_dict)
        else:
            return sorted(ret_list)

    def getGC(self, seq):
        """Get the GC of a sequence"""
        Ns = seq.count('N') + seq.count('n')
        compl = s_maketrans('ACGTacgtnN', '0110011000')
        return sum([float(x) for x in list(seq.translate(compl))])/float(len(seq) - Ns)

    def shiftLowLexi(self, seq):
        """Return the lexicographically lowest form of this sequence"""
        rseq = self.revComp(seq)
        if(seq < rseq):
            return seq
        return rseq

    def shiftLowLexiMer(self, seq):
        """Return the lexicographically lowest form of this kmer"""
        try:
            return self.llDict[seq]
        except KeyError:
            return seq

    def revComp(self, seq):
        """Return the reverse complement of a sequence"""
        # build a dictionary to know what letter to switch to
        return seq.translate(self.compl)[::-1]

    def getKSig(self, seq):
        """Work out kmer signature for a nucleotide sequence

        returns a tuple of floats which is the kmer sig
        """
        # tmp storage
        sig = dict(zip(self.kmerCols, [0.0] * self.numMers))
        # the number fo kmers in this sequence
        num_mers = len(seq)-self.kLen+1
        for i in range(0,num_mers):
            try:
                this_mer = self.llDict[seq[i:i+self.kLen]]
                try:
                    sig[this_mer] += 1.0
                except KeyError:
                    # Ns
                    sig[this_mer] = 1.0
            except KeyError:
                # typically due to an N in the sequence...
                pass

        # normalise by length and return
        return tuple([sig[x] / num_mers for x in self.kmerCols])

###############################################################################
###############################################################################
###############################################################################
###############################################################################
class BamParser:
    """Parse multiple bam files and write the output to hdf5 """

    def __init__(self): pass

    def parse(self, bamFiles, contigNames, cid2Indices):
        """Parse multiple bam files and store the results in the main DB"""
        print "Importing BAM files"
        from bamtyper.utilities import BamParser as BTBP
        BP = BTBP()
        (links, ref_lengths, coverages) = BP.getLinks(bamFiles, full=False, verbose=True, doCoverage=True, minJoin=5)

        # go through all the contigs sorted by name.
        # we want to make an array of tuples of coverage sigs
        cov_sigs = []
        bam_range = range(len(bamFiles))
        for cid in contigNames:
            tmp_cov = []
            for i in bam_range:
                try:
                    tmp_cov.append(coverages[i][cid])
                except KeyError:
                    # may be no coverage for this contig
                    tmp_cov.append(0.0)
            cov_sigs.append(tuple(tmp_cov))

        # transform the links into something a little easier to parse later
        rowwise_links = []
        for cid in links:
            for link in links[cid]:
                try:
                    rowwise_links.append((cid2Indices[cid],          # contig 1
                                          cid2Indices[link[0]],      # contig 2
                                          int(link[1]),               # numReads
                                          int(link[2]),               # linkType
                                          int(link[3])                # gap
                                          ))
                except KeyError:
                    pass
        return (rowwise_links, np.array(cov_sigs))

def getBamDescriptor(fullPath, index_num):
    """AUX: Reduce a full path to just the file name minus extension"""
    return str(index_num) + '_' + op_splitext(op_basename(fullPath))[0]

###############################################################################
###############################################################################
###############################################################################
###############################################################################

class CoverageTransformer:
    """Transform coverage profiles GroopM-style"""

    def __init__(self,
                 numContigs,
                 numStoits,
                 normCoverages,
                 kmerNormPC1,
                 coverageProfiles,
                 stoitColNames,
                 scaleFactor=1000):
        self.numContigs = numContigs
        self.numStoits = numStoits
        self.normCoverages = normCoverages
        self.kmerNormPC1 = kmerNormPC1
        self.covProfiles = coverageProfiles
        self.stoitColNames = stoitColNames
        self.indices = range(self.numContigs)
        self.scaleFactor = scaleFactor
        
        # things we care about!
        self.TCentre = None
        self.transformedCP = np.zeros((self.numContigs,3))
        self.corners = np.zeros((self.numStoits,3))

    def transformCP(self, silent=False, nolog=False):
        """Do the main transformation on the coverage profile data"""
        shrinkFn = np.log10
        if(nolog):
            shrinkFn = lambda x:x

        if(not silent):
            print "    Reticulating splines"
            print "    Dimensionality reduction"

        unit_vectors = [(np.cos(i*2*np.pi/self.numStoits),np.sin(i*2*np.pi/self.numStoits)) for i in range(self.numStoits)]

        # make sure the bams are ordered consistently
        if self.numStoits > 3:
            self.shuffleBAMs()

        for i in range(len(self.indices)):
            shifted_vector = np.array([0.0,0.0])
            try:
                flat_vector = (self.covProfiles[i] / sum(self.covProfiles[i]))
            except FloatingPointError:
                flat_vector = self.covProfiles[i]

            for j in range(self.numStoits):
                shifted_vector[0] += unit_vectors[j][0] * flat_vector[j]
                shifted_vector[1] += unit_vectors[j][1] * flat_vector[j]

            # log scale it towards the centre
            scaling_vector = shifted_vector * self.scaleFactor
            sv_size = np.linalg.norm(scaling_vector)
            if sv_size > 1:
                shifted_vector /= shrinkFn(sv_size)

            self.transformedCP[i,0] = shifted_vector[0]
            self.transformedCP[i,1] = shifted_vector[1]
            # should always work cause we nuked
            # all 0 coverage vecs in parse
            self.transformedCP[i,2] = shrinkFn(self.normCoverages[i])

        # finally scale the matrix to make it equal in all dimensions
        min = np.amin(self.transformedCP, axis=0)
        self.transformedCP -= min
        max = np.amax(self.transformedCP, axis=0)
        max = max / (self.scaleFactor-1)
        self.transformedCP /= max

        # get the corner points
        XYcorners = np.reshape([i for i in np.array(unit_vectors)],
                               (self.numStoits, 2))

        for i in range(self.numStoits):
            self.corners[i,0] = XYcorners[i,0]
            self.corners[i,1] = XYcorners[i,1]

        # shift the corners to match the space
        self.corners -= min
        self.corners /= max

        # scale the corners to fit the plot
        cmin = np.amin(self.corners, axis=0)
        self.corners -= cmin
        cmax = np.amax(self.corners, axis=0)
        cmax = cmax / (self.scaleFactor-1)
        self.corners[:,0] /= cmax[0]
        self.corners[:,1] /= cmax[1]
        for i in range(self.numStoits):
            self.corners[i,2] = self.scaleFactor + 100 # only affect the z axis

        self.TCentre = np.mean(self.corners, axis=0)

    def small2indices(self, index, side):
        """Return the indices of the comparative items
        when given an index into a condensed distance matrix
        """
        step = 0
        while index >= (side-step):
            index = index - side + step
            step += 1
        return (step, step + index + 1)

    def shuffleBAMs(self, ordering=None):
        """Make the data transformation deterministic by reordering the bams"""
        if ordering is None:
            # we will need to deduce the ordering of the contigs
            # first we should make a subset of the total data
            # we'd like to take it down to about 1500 or so RI's
            # but we'd like to do this in a repeatable way
            ideal_contig_num = 1500
            sub_cons = np.arange(self.numContigs)
            while len(sub_cons) > ideal_contig_num:
                # select every second contig when sorted by norm cov
                cov_sorted = np.argsort(self.normCoverages[sub_cons])
                sub_cons = np.array([sub_cons[cov_sorted[i*2]] for i in np.arange(int(len(sub_cons)/2))])
                
                if len(sub_cons) > ideal_contig_num:
                    # select every second contig when sorted by mer PC1
                    mer_sorted = np.argsort(self.kmerNormPC1[sub_cons])
                    sub_cons = np.array([sub_cons[mer_sorted[i*2]] for i in np.arange(int(len(sub_cons)/2))])
    
            # now that we have a subset, calculate the distance between each of the untransformed vectors
            num_sc = len(sub_cons)
            
            # log shift the coverages towards the origin
            sub_covs = np.transpose([self.covProfiles[i]*(np.log10(self.normCoverages[i])/self.normCoverages[i]) for i in sub_cons])
            sq_dists = cdist(sub_covs,sub_covs,'cityblock')
            dists = squareform(sq_dists)
    
            # initialise a list of left, right neighbours
            lr_dict = {}
            for i in range(self.numStoits):
                lr_dict[i] = []
    
            too_big = 10000
            while True:
                closest = np.argmin(dists)
                if dists[closest] == too_big:
                    break
                (i,j) = self.small2indices(closest, self.numStoits-1)
                lr_dict[j].append(i)
                lr_dict[i].append(j)
    
                # mark these guys as neighbours
                if len(lr_dict[i]) == 2:
                    # no more than 2 neighbours
                    sq_dists[i,:] = too_big
                    sq_dists[:,i] = too_big
                    sq_dists[i,i] = 0.0
                if len(lr_dict[j]) == 2:
                    # no more than 2 neighbours
                    sq_dists[j,:] = too_big
                    sq_dists[:,j] = too_big
                    sq_dists[j,j] = 0.0
    
                # fix the dist matrix
                sq_dists[j,i] = too_big
                sq_dists[i,j] = too_big
                dists = squareform(sq_dists)
    
            # now make the ordering
            ordering = [0, lr_dict[0][0]]
            done = 2
            while done < self.numStoits:
                last = ordering[done-1]
                if lr_dict[last][0] == ordering[done-2]:
                    ordering.append(lr_dict[last][1])
                    last = lr_dict[last][1]
                else:
                    ordering.append(lr_dict[last][0])
                    last = lr_dict[last][0]
                done+=1

        # reshuffle the contig order!
        # yay for bubble sort!
        working = np.arange(self.numStoits)
        for i in range(1, self.numStoits):
            # where is this guy in the list
            loc = list(working).index(ordering[i])
            if loc != i:
                # swap the columns
                self.covProfiles[:,[i,loc]] = self.covProfiles[:,[loc,i]]
                self.stoitColNames[[i,loc]] = self.stoitColNames[[loc,i]]
                working[[i,loc]] = working[[loc,i]]



#!/usr/bin/env python
from __future__ import division
###############################################################################
#                                                                             #
# som.py                                                                      #
#                                                                             #
# GroopM self organising map engine                                           #
#                                                                             #
# Copyright (C) Michael Imelfort                                              #
#                                                                             #
###############################################################################
#                                                                             #
#          .d8888b.                                    888b     d888          #
#         d88P  Y88b                                   8888b   d8888          #
#         888    888                                   88888b.d88888          #
#         888        888d888 .d88b.   .d88b.  88888b.  888Y88888P888          #
#         888  88888 888P"  d88""88b d88""88b 888 "88b 888 Y888P 888          #
#         888    888 888    888  888 888  888 888  888 888  Y8P  888          #
#         Y88b  d88P 888    Y88..88P Y88..88P 888 d88P 888   "   888          #
#          "Y8888P88 888     "Y88P"   "Y88P"  88888P"  888       888          #
#                                             888                             #
#                                             888                             #
#                                             888                             #
#                                                                             #
###############################################################################
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the                #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program. If not, see <http://www.gnu.org/licenses/>.        #
#                                                                             #
###############################################################################
# This script uses / is inspired by chunks of code and ideas originally
# developed by Kyle Dickerson and others
# Here: http://www.singingtree.com/~kyle/blog/archives/00000251.htm
# and Here: http://paraschopra.com/sourcecode/SOM/index.php
###############################################################################
## Kyle Dickerson
## kyle.dickerson@gmail.com
## Jan 15, 2008
##
## Self-organizing map using scipy
## This code is licensed and released under the GNU GPL
##
## This code uses a square grid rather than hexagonal grid, as scipy allows for
## fast square grid computation. I designed sompy for speed, so attempting to
## read the code may not be very intuitive. If you're trying to learn how SOMs
## work, I would suggest starting with Paras Chopras SOMPython code:
## http://www.paraschopra.com/sourcecode/SOM/index.php
## It has a more intuitive structure for those unfamiliar with scipy,
## however it is much slower. If you do use this code for something,
## please let me know, I'd like to know if has been useful to anyone.
###############################################################################

__author__ = "Michael Imelfort"
__copyright__ = "Copyright 2012/2013"
__credits__ = ["Michael Imelfort"]
__license__ = "GPL3"
__version__ = "0.1.0"
__maintainer__ = "Michael Imelfort"
__email__ = "mike@mikeimelfort.com"
__status__ = "Beta"

###############################################################################

import sys
from random import randrange, randint, random
from math import log, exp
import numpy as np
from scipy.spatial.distance import cdist
from PIL import Image, ImageDraw
np.seterr(all='raise')

# GroopM imports
from torusMesh import TorusMesh as TM
from rainbow import Rainbow
import groopmExceptions as ge

###############################################################################
###############################################################################
###############################################################################
###############################################################################

class SOM:
    """A single instance of a self organising map"""
    def __init__(self, side, dimension, lc=None, uc=None):
        self.side = side # side length of the grid
        self.dimension = dimension # size of our grid vectors
        self.bestMatchCoords = [] # x/y coords of classified vectors

        self.radius = float(side)/2 # the radius of neighbour nodes which will be influenced by each new training vector

        # initialise the nodes to random values between 0 -> 1
        self.weights = TM(self.side, dimension=self.dimension, randomize=True)

        # cutoff for turning the VS_flat into a boundary mask
        # this is a magic number, but it seems to work OK
        self.maskCutoff = 16
        self.boundaryMask = np.zeros((self.side,self.side))
        self.VS_flat = np.zeros((self.side,self.side))

        # bin assignments
        self.binAssignments = np.zeros((self.side,self.side)) # the 0 bin is 'not assigned'

        if lc is not None:
            diff = uc - lc
            self.weights.nodes *= uc
            self.weights.nodes += lc
        self.regions = None

        # we'd like to know who is next to whom
        self.regionNeighbours = {}

    def getWeights(self):
        """Get the weights nodes"""
        return self.weights.nodes

    def getRegions(self):
        """Get the regions nodes"""
        return self.regions.nodes

    def loadWeights(self, nodes):
        """Use externally supplied data"""
        self.weights.nodes = nodes

    def loadRegions(self, nodes):
        """Use externally supplied regions"""
        self.regions = TM(self.side, dimension=1)
        self.regions.nodes = nodes

#------------------------------------------------------------------------------
# CLASSIFICATION

    # Classify!
    # run this after training to get an X/Y for each vector
    # you'd like to classify
    def classify(self, point):
        """Classify an individual point"""
        (row,col) = self.weights.bestMatch(point)
        return self.regions.nodes[row,col][0]

    def regionalise(self, bids, trainVector):
        """Create regions on the torus based on matches with the training vector
        Train vector is a list of numpy arrays
        """
        # build a regions structure
        self.regions = TM(self.side, dimension=1)
        for row in range(self.side):
            for col in range(self.side):
                self.regions.nodes[row,col] = self.classifyPoint(self.weights.nodes[row,col],
                                                                 trainVector,
                                                                 bids)

    def classifyPoint(self, point, trainVector, bids):
        """Returns the bid of the best match to the trainVector
        trainVector and bids must be in sync
        """
        return bids[np.argmin((((trainVector - point)**2).sum(axis=1))**0.5)]

    def findRegionNeighbours(self):
        """Find out which regions neighbour which other regions"""
        neighbours = {}
        self.regionNeighbours = {}
        for row in range(self.side-1):
            for col in range(self.side-1):
                s_bid = self.regions.nodes[row,col][0]
                # test against right, down and diagonally down
                q_bids = [self.regions.nodes[row,col+1][0],
                          self.regions.nodes[row+1,col][0],
                          self.regions.nodes[row+1,col+1][0]
                         ]
                for q_bid in q_bids:
                    if(s_bid != q_bid):
                        # make storage for this region
                        if(s_bid not in self.regionNeighbours):
                            self.regionNeighbours[s_bid] = []
                        if(q_bid not in self.regionNeighbours):
                            self.regionNeighbours[q_bid] = []
                        # add in the neighbours
                        if(q_bid not in self.regionNeighbours[s_bid]):
                            self.regionNeighbours[s_bid].append(q_bid)
                        if(s_bid not in self.regionNeighbours[q_bid]):
                            self.regionNeighbours[q_bid].append(s_bid)

                        # we only need to return a tuple
                        nt = self.makeNTuple(s_bid,q_bid)
                        neighbours[nt] = True
        return neighbours.keys()

    def makeNTuple(self, bid1, bid2):
        """A way for making standard tuples from bids"""
        if(bid1 < bid2): return (bid1, bid2)
        return (bid2, bid1)

    def getNeighbours(self, bids):
        """return the neighbours of these bids"""
        ret_list = []
        for bid in bids:
            if(bid in self.regionNeighbours):
                ret_list.extend([i for i in self.regionNeighbours[bid] if i not in ret_list])
        return ret_list

#------------------------------------------------------------------------------
# TRAINING

    def train(self,
              trainVector,
              weights=None,
              iterations=1000,
              vectorSubSet=1000,
              weightImgFileNamePrefix="",
              epsilom=0.0001,
              influenceRate=0.4,
              mask=None,
              radius=0.,
              silent=True
              ):
        """Train the SOM
        Train vector is a list of numpy arrays
        """

        if not silent:
            print "    Start SOM training. Side: %d Max: %d iterations" % (self.side, iterations)

        if radius == 0.0:
            radius = self.radius

        # we can use a dummy set of weights, or the *true* weights
        if weights is None:
            replace_weights = True
            weights = self.weights.nodes
            flat_nodes = self.weights.flatNodes
            rows = self.side
            cols = self.side
        else:
            shape = np.shape(weights)
            rows = shape[0]
            cols = shape[1]
            flat_nodes = weights.reshape((rows*cols, self.dimension))
            replace_weights = False

        # over time we'll shrink the radius of nodes which
        # are influenced by the current training node
        time_constant = iterations/log(radius)

        # we would ideally like to select guys from the training set at random
        if(len(trainVector) <= vectorSubSet):
            index_array = np.arange(len(trainVector))
            cut_off = len(trainVector) # if less than 1000 training vectors, set this to suit
        else:
            rand_index_array = np.arange(len(trainVector))
            cut_off = vectorSubSet

        for i in range(1, iterations+1):
            if not silent:
                sys.stdout.write("\r    Iteration: % 4d of % 4d" % (i, iterations))
                sys.stdout.flush()

#--------
# Make stamp
            # gaussian decay on radius and amount of influence
            radius_decaying=radius*exp(-1.0*i/time_constant)
            if(radius_decaying < 2):
                return weights

            grad = -1 * influenceRate / radius_decaying
            # we will make a "stamp" to speed things up
            max_radius = int(radius_decaying)
            q_stamp = np.zeros((max_radius+1,max_radius+1))

            for row in range(0, max_radius+1):
                for col in range(0, max_radius + 1):
                    # now we check to see that the euclidean distance is less than
                    # the specified distance.
                    true_dist = np.sqrt( row**2 + col**2 )
                    #if true_dist > 0.0:
                    check_dist = np.round(true_dist+0.00001)
                    if(check_dist <= radius_decaying):
                        # influence is propotional to distance
                        influence = true_dist*grad + influenceRate
                        q_stamp[row, col] = influence
                        q_stamp[col, row] = influence

            # divide by 2, so we don't mess up the stamp
            q_stamp[:,0] /= 2
            q_stamp[0,:] /= 2
            stamp = np.zeros((2*max_radius+1,2*max_radius+1))
            # bottom right
            stamp[max_radius:,max_radius:] += q_stamp
            # top right
            stamp[:max_radius+1,max_radius:] += np.rot90(q_stamp,1)
            # top left
            stamp[:max_radius+1,:max_radius+1] += np.rot90(q_stamp,2)
            # bottom left
            stamp[max_radius:,:max_radius+1] += np.rot90(q_stamp,3)
            # center
            stamp[max_radius, max_radius] = influenceRate


            # now find where the useless info is and cull it from the stamp
            max_vals = np.max(stamp, axis=0)
            k = 0
            while k < len(max_vals):
                if max_vals[k] > 0.0:
                    break
                k += 1
            if k < len(max_vals):
                stamp = stamp[k:2*max_radius+1-k,k:2*max_radius+1-k]

            # keep track of how big the stamp is now
            stamp_side = len(stamp)
            stamp_radius = int((stamp_side-1)/2)

            # if there are more than vectorSubSet training vecs
            # take a random selection
            if(len(trainVector) > vectorSubSet):
                np.random.shuffle(rand_index_array)
                index_array = rand_index_array[:cut_off]
#--------
# Make worksheet
            worksheet = np.zeros(self.dimension*rows*cols*9).reshape((rows*3,
                                                                      cols*3,
                                                                      self.dimension))
            worksheet[0:rows,0:cols] = weights
            worksheet[0:rows,cols:cols*2] = weights
            worksheet[0:rows,cols*2:cols*3] = weights
            worksheet[rows:rows*2,0:cols] = weights
            worksheet[rows:rows*2,cols:cols*2] = weights
            worksheet[rows:rows*2,cols*2:cols*3] = weights
            worksheet[rows*2:rows*3,0:cols] = weights
            worksheet[rows*2:rows*3,cols:cols*2] = weights
            worksheet[rows*2:rows*3,cols*2:cols*3] = weights

            # make a set of "delta nodes"
            # these contain the changes to the set of grid nodes
            # and we will add their values to the grid nodes
            # once we have input all the training nodes
            deltasheet = np.zeros_like(worksheet)

            for j in index_array:
                # find the best match between then training vector and the
                # current grid, inlined for greater speed
                loc = np.argmin(cdist(flat_nodes, [trainVector[j]]))
                row = int(loc/cols)
                col = loc-(row*cols)

                # row col represent the center of the stamp
                weights_patch = worksheet[rows+row-stamp_radius:rows+row+stamp_radius+1,
                                          cols+col-stamp_radius:cols+col+stamp_radius+1]
                weights_patch = -1*(weights_patch - trainVector[j])
                #print row, col, rows, cols, stamp_radius, np.shape(weights_patch), np.shape(weights_patch[:,:,0]), np.shape(stamp), np.shape(deltasheet[rows+row-stamp_radius:rows+row+stamp_radius+1,cols+col-stamp_radius:cols+col+stamp_radius+1])
                weights_patch[:,:,0] *= stamp
                weights_patch[:,:,1] *= stamp
                weights_patch[:,:,2] *= stamp
                weights_patch[:,:,3] *= stamp

                deltasheet[rows+row-stamp_radius:rows+row+stamp_radius+1,
                           cols+col-stamp_radius:cols+col+stamp_radius+1] += weights_patch

            # now fold the deltas and update the weights
            deltasheet[:,cols:2*cols] += deltasheet[:,0:cols]
            deltasheet[:,cols:2*cols] += deltasheet[:,2*cols:3*cols]
            deltasheet[rows:2*rows,cols:2*cols] += deltasheet[0:rows,cols:2*cols]
            deltasheet[rows:2*rows,cols:2*cols] += deltasheet[2*rows:3*rows,cols:2*cols]

            # add the deltas to the grid nodes and clip to keep between 0 and 1
            if mask is None:
                weights = np.clip(weights + deltasheet[rows:2*rows,cols:2*cols], 0, 1)
            else:
                delta_fold = deltasheet[rows:2*rows,cols:2*cols]
                for (r,c) in mask.keys():
                    weights[r,c] = np.clip(weights[r,c] + delta_fold[r,c], 0, 1)
                    flat_nodes = weights.reshape((rows*cols, self.dimension))

            if replace_weights == True:
                flat_nodes = self.weights.fixFlatNodes(weights=weights)

            # make a tmp image, perhaps
            if(weightImgFileNamePrefix != ""):
                filename = "%s_%04d.jpg" % (weightImgFileNamePrefix, i)
                print " writing: %s" % filename
                self.weights.renderSurface(filename)

        return weights

    def makeBoundaryMask(self, plotMaskFile=""):
        """Make a mask for cutting out boundaries"""
        # First create the mask
        VS = self.weights.buildVarianceSurface()
#        self.VS_flat = np.array([[int(j) for j in i] for i in np.array(VS[:,:,0] + VS[:,:,1] + VS[:,:,2] + VS[:,:,3])*250]).reshape((self.side, self.side))
        self.VS_flat = np.array(VS[:,:,0] + VS[:,:,1] + VS[:,:,2] + VS[:,:,3]).reshape((self.side, self.side)) * 250
        self.boundaryMask = np.where(self.VS_flat > self.maskCutoff, 1., 0.)

        if plotMaskFile != "":
            self.renderBoundaryMask(plotMaskFile)

    def maskBoundaries(self, addNoise=False, weights=None, mask=None, doFlat=False):
        """mask boundaries and add some random noise to
        some non-masked areas if asked to"""
        max_noise = 0.1
        noise_targets = 3
        if weights is None:
            weights = self.weights.nodes
            rows = self.side
            cols = self.side
        else:
            shape = np.shape(weights)
            rows = shape[0]
            cols = shape[1]

        if mask is None:
            mask = self.boundaryMask
        for r in range(rows):
            for c in range(cols):
                if mask[r,c] == 1:
                    # on the boundary, mask as -1's
                    weights[r,c] = [-1.]*self.dimension
                elif addNoise:
                    if randint(10) <= noise_targets:
                        # add some noise
                        noise_amount = random() * max_noise + 1.0
                        weights[r,c] *= noise_amount
        if doFlat:
            self.weights.fixFlatNodes()

    def defineBinRegions(self, bids, binProfiles, render=False):
        """Work out which bins go where"""
        rcols = {}
        rand_col_lower = 15
        rand_col_upper = 200
        bp_map = {}

        # use a flood fill algorithm to color in adjacent spots
        # and assign bins to unmasked points
        for i in range(len(bids)):
            # find out where this bin matches bestest
            [row, col] = self.weights.bestMatch(binProfiles[i])
            bid = bids[i]
            bp_map[bid] = binProfiles[i]
            rcols[bid] = (randrange(rand_col_lower, rand_col_upper),
                          randrange(rand_col_lower, rand_col_upper),
                          randrange(rand_col_lower, rand_col_upper)
                         )
            self.expandAssign(row, col, bid, bp_map)
        if render:
            self.renderBoundaryMask("S3.png", colMap=rcols)

        # now clean up the mask
        for r in range(self.side):
            for c in range(self.side):
                if self.boundaryMask[r,c] == 0 and self.binAssignments[r,c] == 0:
                    # unmasked AND unassigned
                    self.boundaryMask[r,c] = 1
        if render:
            self.renderBoundaryMask("S4.png", colMap=rcols)

    def expandAssign(self, startR, startC, bid, binProfileMap):
        """Based on floodfill, add more points to a bin assignment"""
        # get all the points within this region
        points = self.floodFill(startR, startC, self.boundaryMask)
        collision_bid = 0
        for (r,c) in points.keys():
            if self.binAssignments[r,c] != 0:
                if self.binAssignments[r,c] != bid:
                    # we have already assigned this point to a bin
                    # most likely we need to set the mask cutoff higher, but just for this region
                    collision_bid = self.binAssignments[r,c]
                    #print "\n", (r,c), "already assigned to bin %d, trying to reassign to bin %d" % (collision_bid, bid)
                    re_calc_mask = True
                    break
            self.binAssignments[r,c] = bid

        if collision_bid != 0:
            resolved = False
            [crow, ccol] = self.weights.bestMatch(binProfileMap[collision_bid])   # where the old bin's floodfill started
            mc = self.maskCutoff
            # we can't do anything if we can't lower the cutoff...
            while mc >= 2:
                # rebuild the mask with a new cutoff
                mc = mc/2
                mask = np.copy(self.boundaryMask)
                for (r,c) in points.keys():
                    if self.VS_flat[r,c] > mc:
                        mask[r,c] = 1.
                    else:
                        mask[r,c] = 0.
                #self.renderBoundaryMask("MASK_%d_%d_%f.png" % (bid, collision_bid, mc), mask)
                collision_points = self.floodFill(crow, ccol, mask)
                new_points = self.floodFill(startR, startC, mask)
                #print len(collision_points.keys()), len(new_points.keys())
                #print collision_points.keys()[0] in new_points
                if len(collision_points.keys()) == 0 or len(new_points.keys()) == 0:
                    continue
                # there should be no overlap
                if collision_points.keys()[0] not in new_points:
                    # we have resolved the issue
                    resolved = True
                    # now we need to fix the binAssignments and boundary mask
                    self.boundaryMask = mask
                    for (r,c) in points.keys():
                        if (r,c) in new_points:
                            # assign this point to the new bid
                            self.binAssignments[r,c] = bid
                        elif (r,c) in collision_points:
                            # point in the new mask
                            self.binAssignments[r,c] = collision_bid
                        else:
                            self.binAssignments[r,c] = 0.
                    break

            if not resolved:
                print "Cannot repair map, bin %d may be incorrectly merged with bin %d" % (bid, collision_bid)
                return

    def makeBinMask(self, profile, fileName="", dim=False):
        """Return a mask of the region this profile falls in"""
        [r, c] = self.weights.bestMatch(profile)
        points = self.floodFill(r, c, self.boundaryMask)
        if fileName != "":
            ret_mask = np.ones_like(self.boundaryMask)
            for (r,c) in points.keys():
                ret_mask[r,c] = 0
            self.renderBoundaryMask(fileName, mask=ret_mask)

        return points

    def floodFill(self, startR, startC, mask):
        """Return all points affected by a flood fill operation at the given point"""
        points = {}
        toFill = set()
        toFill.add((startR, startC))
        seen = {(startR, startC) : True}
        while len(toFill) != 0:
            (r,c) = toFill.pop()
            if mask[r,c] == 1:
                # we are at the boundary of a region
                continue

            points[(r,c)] = [r,c]

            # don't forget we're on a torus
            if r == 0: rm1 = self.side - 1
            else: rm1 = r - 1
            if c == 0: cm1 = self.side - 1
            else: cm1 = c - 1
            if r == self.side - 1: rp1 = 0
            else: rp1 = r + 1
            if c == self.side - 1:cp1 = 0
            else: cp1 = c + 1
            if (rm1,c) not in seen: toFill.add((rm1,c)); seen[(rm1,c)] = True
            if (rp1,c) not in seen: toFill.add((rp1,c)); seen[(rp1,c)] = True
            if (r,cm1) not in seen: toFill.add((r,cm1)); seen[(r,cm1)] = True
            if (r,cp1) not in seen: toFill.add((r,cp1)); seen[(r,cp1)] = True

        return points

    def secondsToStr(self, t):
        rediv = lambda ll,b : list(divmod(ll[0],b)) + ll[1:]
        return "%d:%02d:%02d.%03d" % tuple(reduce(rediv,[[t*1000,],1000,60,60]))

#------------------------------------------------------------------------------
# CLASSIFICATION

    def classifyContig(self, profile):
        """Classify this contig"""
        [r,c] = self.weights.bestMatch(profile)
        return int(self.binAssignments[r,c])

#------------------------------------------------------------------------------
# IO and IMAGE RENDERING

    def renderWeights(self, tag, weights=None):
        """Render the surface weights"""
        filename = tag+".png"
        if weights is None:
            self.weights.renderSurface(filename)
        else:
            self.weights.renderSurface(filename, nodes=weights)

    def renderRegions(self, tag, palette):
        """Render the regions
        palette is a hash of bid -> color
        """
        filename = tag+".png"
        if(self.regions is None):
            raise ge.RegionsDontExistException
        try:
            img = Image.new("RGB", (self.weights.rows,self.regions.columns))
            for row in range(self.side):
                for col in range(self.side):
                    img.putpixel((col,row), palette[self.regions.nodes[row,col][0]])
            img = img.resize((self.weights.columns*10, self.weights.rows*10),Image.NEAREST)
            img.save(filename)
        except:
            print sys.exc_info()[0]
            raise

    def renderBoundaryMask(self, fileName, mask=None, colMap=None):
        """Plot the boundary mask"""
        if mask is None:
            mask = self.boundaryMask
        try:
            img = Image.new("RGB", (self.side, self.side))
            for r in range(self.side):
                for c in range(self.side):
                    if mask[r,c] == 0:
                        if colMap is not None:
                            try:
                                col = colMap[self.binAssignments[r,c]]
                            except KeyError:
                                col = (255,255,255)
                        else:
                            col = (255,255,255)
                        img.putpixel((c,r), col)
                    else:
                        img.putpixel((c,r), (0,0,0))
            img = img.resize((self.side*10, self.side*10),Image.NEAREST)
            img.save(fileName)
        except:
            print sys.exc_info()[0]
            raise

    def transColour(self, val):
        """Transform color value"""
        return 10 * log(val)

    def renderBestMatches(self, fileName, weighted=False):
        """make an image of where best matches lie
        set weighted to use a heatmap view of where they map
        """
        img_points = np.zeros((self.weights.rows,self.weights.columns))
        try:
            img = Image.new("RGB", (self.weights.columns, self.weights.rows))
            if(weighted): # color points by bestmatch density
                max = 0
                for point in self.bestMatchCoords:
                    img_points[point[0],point[1]] += 1
                    if(max < img_points[point[0],point[1]]):
                        max = img_points[point[0],point[1]]
                max += 1
                resolution = 200
                if(max < resolution):
                    resolution = max - 1
                max = self.transColour(max)
                rainbow = Rainbow(0, max, resolution, "gbr")
                for point in self.bestMatchCoords:
                    img.putpixel((point[1],point[0]), rainbow.getColour(self.transColour(img_points[point[0],point[1]])))
            else: # make all best match points white
                for point in self.bestMatchCoords:
                    img.putpixel((point[1],point[0]), (255,255,255))
            img = img.resize((self.weights.columns*10, self.weights.rows*10),Image.NEAREST)
            img.save(fileName)
        except:
            print sys.exc_info()[0]
            raise

###############################################################################
###############################################################################
###############################################################################
###############################################################################



#!/usr/bin/env python
###############################################################################
#                                                                             #
# torusMesh.py                                                                #
#                                                                             #
# Implements structure and methods for working with a torus shaped mesh       #
#                                                                             #
# Copyright (C) Michael Imelfort                                              #
#                                                                             #
###############################################################################
#                                                                             #
#          .d8888b.                                    888b     d888          #
#         d88P  Y88b                                   8888b   d8888          #
#         888    888                                   88888b.d88888          #
#         888        888d888 .d88b.   .d88b.  88888b.  888Y88888P888          #
#         888  88888 888P"  d88""88b d88""88b 888 "88b 888 Y888P 888          #
#         888    888 888    888  888 888  888 888  888 888  Y8P  888          #
#         Y88b  d88P 888    Y88..88P Y88..88P 888 d88P 888   "   888          #
#          "Y8888P88 888     "Y88P"   "Y88P"  88888P"  888       888          #
#                                             888                             #
#                                             888                             #
#                                             888                             #
#                                                                             #
###############################################################################
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the                #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program. If not, see <http://www.gnu.org/licenses/>.        #
#                                                                             #
###############################################################################

__author__ = "Michael Imelfort"
__copyright__ = "Copyright 2012/2013"
__credits__ = ["Michael Imelfort"]
__license__ = "GPL3"
__version__ = "0.1.0"
__maintainer__ = "Michael Imelfort"
__email__ = "mike@mikeimelfort.com"
__status__ = "Beta"

###############################################################################
import sys
import numpy as np
from PIL import Image, ImageDraw
from scipy.spatial.distance import cdist
from colorsys import hsv_to_rgb as htr

np.seterr(all='raise')

###############################################################################
###############################################################################
###############################################################################
###############################################################################

class UnknownDistanceType(BaseException):
    pass

###############################################################################
###############################################################################
###############################################################################
###############################################################################

class TorusMesh:
    """A closed mesh, in the shape of a torus"""
    
    def __init__(self, rows, columns=0, dimension=1, randomize=False):
        """ init
        Set columns if you'd like something other than a square
        By default the torus is a scalar field.
        Increase dimension if you'd like a vector field
        """
        self.rows = rows
        if(columns == 0): # make it square
            self.columns = rows
        else:
            self.columns = columns
        self.dimension = dimension
        self.shape = (self.rows,self.columns,self.dimension)
        self.flatShape = (self.rows*self.columns,self.dimension)
        self.size = self.dimension*self.rows*self.columns
        
        # we use these values in dist many many times
        self.halfRow = float(self.rows)/2
        self.halfColumn = float(self.columns)/2
        
        # the first two dimensions repesent points on the surface
        # the remainder represent values
        if(randomize):
            self.nodes = np.random.random(self.size).reshape(self.shape)
        else:
            self.nodes = np.zeros(self.size).reshape(self.shape)

        # make an array of flattened nodes
        self.flatNodes = self.nodes.reshape(self.flatShape)
        
        # work out the central vector and corresponding max angle
        c_vec = np.ones((self.dimension))
        self.largestMag = np.linalg.norm(c_vec)
        self.cVec = c_vec / self.largestMag 
        top_vec = np.zeros_like(self.cVec)
        top_vec[0] = 1
        self.maxAngle = self.getAngBetweenNormed(top_vec, self.cVec)

    def fixFlatNodes(self, weights=None):
        """Make sure everything is in sync"""
        if weights is not None:
            self.nodes = weights
        self.flatNodes = self.nodes.reshape(self.flatShape)
        return self.flatNodes     
            
#------------------------------------------------------------------------------
# WORKING WITH THE DATA

    def bestMatch(self, targetVector):
        """Returns location of the best match to an existing vector
        uses Euclidean distance
        """
        loc = np.argmin(cdist(self.flatNodes, [targetVector]))
        row = int(loc/self.columns)
        col = loc-(row*self.rows)
        return [row, col]

    def buildVarianceSurface(self):
        """Work out the difference between each point and it's eight neighbours"""
        diff_array = np.zeros(self.shape)
        shift_array = np.zeros(self.shape)
        shift_array2 = np.zeros(self.shape)
        shift_diff = np.zeros(self.shape)

        # shift horizontal
        # ---
        # BA-
        # ---
        shift_array[:,:-1,:] = self.nodes[:,1:self.columns,:]   # shift left
        shift_array[:,-1,:] = self.nodes[:,0,:]                 # first node col to the end
        tmp_diff = np.abs(shift_array-self.nodes)           # get the difference
        diff_array += tmp_diff                              # add it on
        shift_diff[:,0,:] = tmp_diff[:,-1,:]                    # shift diff right
        shift_diff[:,1:self.columns,:] = tmp_diff[:,:-1,:]      # last to the first
        diff_array += shift_diff                            # add it on

        # shift horizontal vertical
        # B--
        # -A-
        # ---
        shift_array2[:-1,:,:] = shift_array[1:self.columns,:,:]
        shift_array2[-1,:,:] = shift_array[0,:,:]                
        tmp_diff = np.abs(shift_array2-self.nodes)          
        diff_array += tmp_diff                             
        shift_diff[0,:,:] = tmp_diff[-1,:,:]                   
        shift_diff[1:self.columns,:,:] = tmp_diff[:-1,:,:]     
        tmp_diff[:,0,:] = shift_diff[:,-1,:]
        tmp_diff[:,1:self.columns,:] = shift_diff[:,:-1,:]
        diff_array += tmp_diff

        # shift vertical
        # -B-
        # -A-
        # ---
        shift_array[:-1,:,:] = self.nodes[1:self.columns,:,:]
        shift_array[-1,:,:] = self.nodes[0,:,:]                
        tmp_diff = np.abs(shift_array-self.nodes)          
        diff_array += tmp_diff                             
        shift_diff[0,:,:] = tmp_diff[-1,:,:]                   
        shift_diff[1:self.columns,:,:] = tmp_diff[:-1,:,:]     
        diff_array += shift_diff                           

        # shift vertical horizontal
        # --B
        # -A-
        # ---
        shift_array2[:,0,:] = shift_array[:,-1,:]
        shift_array2[:,1:self.columns,:] = shift_array[:,:-1,:]
        tmp_diff = np.abs(shift_array2-self.nodes)
        diff_array += tmp_diff
        shift_diff[:,:-1,:] = tmp_diff[:,1:self.columns,:]
        shift_diff[:,-1,:] = tmp_diff[:,0,:]
        tmp_diff[0,:,:] = shift_diff[-1,:,:]                   
        tmp_diff[1:self.columns,:,:] = shift_diff[:-1,:,:]     
        diff_array += tmp_diff

        return diff_array
    
#------------------------------------------------------------------------------
# COLORING
    
    def getColor(self, vector):
        """return a colour for a given weight vector"""
        sn = np.linalg.norm(vector)
        if sn > 0:
            vv = vector / sn
            ang_perc = self.getAngBetweenNormed(vv, self.cVec)/self.maxAngle
            mag_perc = sn / self.largestMag
        else:
            ang_perc = 0.0
            mag_perc = 0.0
        V = 1       # VAL remain fixed at 1. Reduce to make pastels if that's your preference...
        col = [int(i*255) for i in htr(ang_perc, mag_perc, V)]
        return col

    def getAngBetweenNormed(self, P1, P2):
        """Return the angle between two points (in radians)"""
        # find the existing angle between them theta
        c = np.dot(P1,P2)
        # rounding errors hurt everyone...
        if(c > 1.0):
            c = 1.0
        elif(c < -1.0):
            c = -1.0
        return np.arccos(c) # in radians

#------------------------------------------------------------------------------
# IO and IMAGE RENDERING

    def __str__(self):
        """string method"""
        ret_array = []
        for r in range(self.rows):
            for c in range(self.columns):
                ret_array.append("[ ")
                for v in range(self.dimension):
                    ret_array.append(str(self.nodes[r,c,v])+" ")
                ret_array.append("],")
            ret_array.append("\n")
        return "".join(ret_array)

    def renderSurface(self, fileName, nodes=None):
        """make an image of the weights in the som"""
        if nodes is None:
            nodes = self.nodes
        ns = np.shape(nodes)
        rows = ns[0]
        columns = ns[1]
        try:
            img = Image.new("RGB", (columns, rows))
            for r in range(rows):
                # build a color value for a vector value
                for c in range(columns):
                    col = self.getColor(nodes[r,c])
                    img.putpixel((c,r), (col[0], col[1], col[2]))
            img = img.resize((columns*10, rows*10),Image.NEAREST)
            img.save(fileName)
        except:
            print sys.exc_info()[0]
            raise
        
###############################################################################
###############################################################################
###############################################################################
###############################################################################



#!/usr/bin/env python
###############################################################################
#                                                                             #
#    groopmUtils.py                                                           #
#                                                                             #
#    Classes for non-clustering data manipulation and output                  #
#                                                                             #
#    Copyright (C) Michael Imelfort                                           #
#                                                                             #
###############################################################################
#                                                                             #
#          .d8888b.                                    888b     d888          #
#         d88P  Y88b                                   8888b   d8888          #
#         888    888                                   88888b.d88888          #
#         888        888d888 .d88b.   .d88b.  88888b.  888Y88888P888          #
#         888  88888 888P"  d88""88b d88""88b 888 "88b 888 Y888P 888          #
#         888    888 888    888  888 888  888 888  888 888  Y8P  888          #
#         Y88b  d88P 888    Y88..88P Y88..88P 888 d88P 888   "   888          #
#          "Y8888P88 888     "Y88P"   "Y88P"  88888P"  888       888          #
#                                             888                             #
#                                             888                             #
#                                             888                             #
#                                                                             #
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = "Michael Imelfort"
__copyright__ = "Copyright 2012/2013"
__credits__ = ["Michael Imelfort"]
__license__ = "GPL3"
__version__ = "0.2.5"
__maintainer__ = "Michael Imelfort"
__email__ = "mike@mikeimelfort.com"
__status__ = "Beta"

###############################################################################
import os
import sys
import errno

import matplotlib.pyplot as plt
from colorsys import hsv_to_rgb as htr

import numpy as np

# GroopM imports
import binManager
import mstore

# other local imports
from bamtyper.utilities import BamParser as BTBP
np.seterr(all='raise')

###############################################################################
###############################################################################
###############################################################################
###############################################################################

class GMExtractor:
    """Used for extracting reads and contigs based on bin assignments"""
    def __init__(self, dbFileName,
                 bids=[],
                 folder='',
                 ):
        self.dbFileName = dbFileName

        if bids is None:
            self.bids = []
        else:
            self.bids = bids

        if(folder == ''):
            # write to current working dir
            self.outDir = os.getcwd()
        else:
            self.outDir = folder

        # make the dir if need be
        makeSurePathExists(self.outDir)

    def extractContigs(self, timer, fasta=[], cutoff=0):
        """Extract contigs and write to file"""
        self.BM = binManager.BinManager(dbFileName=self.dbFileName)   # bins
        self.BM.loadBins(timer, makeBins=True,silent=False,bids=self.bids)
        self.PM = self.BM.PM

        # load all the contigs which have been assigned to bins
        CP = mstore.ContigParser()
        # contigs looks like cid->seq
        contigs = {}
        try:
            for file_name in fasta:
                with open(file_name, "r") as f:
                    contigs = CP.getWantedSeqs(f, self.PM.contigNames, storage=contigs)
        except:
            print "Could not parse contig file:",fasta[0],sys.exc_info()[0]
            raise

        # now print out the sequences
        print "Writing files"
        for bid in self.BM.getBids():
            if self.BM.PM.isLikelyChimeric[bid]:
                file_name = os.path.join(self.outDir, "BIN_%d.chimeric.fa" % bid)
            else:
                file_name = os.path.join(self.outDir, "BIN_%d.fa" % bid)
            try:
                with open(file_name, 'w') as f:
                    for row_index in self.BM.getBin(bid).rowIndices:
                        cid = self.PM.contigNames[row_index]
                        if(cid in contigs):
                            f.write(">%s\n%s\n" % (cid, contigs[cid]))
                        else:
                            print "WTF", bid, cid
            except:
                print "Could not open file for writing:",file_name,sys.exc_info()[0]
                raise

    def  extractReads(self, timer, bams=[]):
        """Extract reads from sam files and write to file"""
        # load data
        self.BM = binManager.BinManager(dbFileName=self.dbFileName)   # bins
        self.BM.loadBins(timer, makeBins=True,silent=False,bids=self.bids)
        self.PM = self.BM.PM

        print "Extracting reads"

        # work out a set of targets to pass to the parser
        targets = {}
        bids = self.BM.getBids()
        for bid in bids:
            bin = self.BM.getBin(bid)
            for row_index in bin.rowIndices:
                targets[self.PM.contigNames[row_index]] = bid

        # get something to parse the bams with
        bam_parser = BTBP()
        bam_parser.extractReads(bams,
                                '',
                                targets,
                                combineBams=False,
                                headersOnly = True,
                                dontTrustSamFlags=False,
                                folder=self.outDir,
                                verbose=True
                                )

def makeSurePathExists(path):
    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise


###############################################################################
###############################################################################
###############################################################################
###############################################################################

class BinExplorer:
    """Inspect bins, used for validation"""
    def __init__(self,
                 dbFileName,
                 bids=[],
                 transform=True,
                 cmstring="HSV",
                 ignoreContigLengths=False):
        
        self.ignoreContigLengths = ignoreContigLengths 
        self.transform = transform
        self.cmString = cmstring
        self.BM = binManager.BinManager(dbFileName=dbFileName)   # bins
        self.PM = self.BM.PM
        self.PM2 = None
        if bids is None:
            self.bids = []
        else:
            self.bids = bids

    def plotHighlights(self, timer, bids, elevation, azimuth, file, filetype, dpi, alpha, invert=False, show=False):
        """Plot a high def image suitable for publication"""
        self.BM.loadBins(timer,
                         makeBins=True,
                         silent=False,
                         loadContigLengths=True,
                         loadContigNames=False,
                         transform = self.transform)
        if len(self.BM.bins) == 0:
            print "Sorry, no bins to plot"
        else:
            print "Plotting image"
            self.BM.setColorMap(self.cmString)
            fig = plt.figure()
            bins=[]
            if bids is not None:
                for bid in bids:
                    bins.append(self.BM.getBin(bid))

            if show:
                file=""
            elif not file.endswith(filetype):
                file += "." + filetype

            self.PM.renderTransCPData(fileName=file,
                                      elev=elevation,
                                      azim=azimuth,
                                      primaryWidth=6,
                                      dpi=dpi,
                                      showAxis=True,
                                      format=filetype,
                                      fig=fig,
                                      highlight=bins,
                                      alpha=alpha
                                      )
            del fig

            if invert:
                # invert the colors
                from PIL import Image
                import PIL.ImageOps
                image = Image.open(file)
                inverted_image = PIL.ImageOps.invert(image)
                inverted_image.save(file)

    def plotFlyOver(self, timer, fps=10.0, totalTime=120.0):
        """Plot a flyover of the data with bins being removed"""
        self.BM.loadBins(timer,
                         makeBins=True,
                         silent=False,
                         bids=self.bids,
                         loadContigLengths=False,
                         loadContigNames=False,
                         transform = self.transform)
        if len(self.BM.bins) == 0:
            print "Sorry, no bins to plot"
        else:
            print "Plotting flyover"
            self.BM.setColorMap(self.cmString)
            all_bids = self.BM.getBids()

            # control image form and output
            current_azim = 45.0
            current_elev = 0.0
            current_frame = 0.0
            total_frames = fps * totalTime
            total_azim_shift = 720.0
            total_elev_shift = 360.0
            azim_increment = total_azim_shift / total_frames
            elev_increment = total_elev_shift / total_frames

            print "Need",total_frames,"frames:"
            # we need to know when to remove each bin
            bid_remove_rate = float(len(all_bids)) / total_frames
            bids_removed = 0
            current_bid_index = 0
            current_bid = all_bids[current_bid_index]
            restricted_bids = []
            fig = plt.figure()
            while(current_frame < total_frames):
                print "Frame",int(current_frame)
                file_name = "%04d" % current_frame +".jpg"
                self.PM.renderTransCPData(fileName=file_name,
                                             elev=current_elev,
                                             azim=current_azim,
                                             primaryWidth=6,
                                             dpi=200,
                                             showAxis=True,
                                             format='jpeg',
                                             fig=fig,
                                             restrictedBids = restricted_bids
                                             )
                current_frame += 1
                current_azim += azim_increment
                current_elev += elev_increment
                print bid_remove_rate*current_frame, current_frame, "BR:",bids_removed, int(bid_remove_rate*current_frame)
                while bids_removed < int(bid_remove_rate*current_frame):
                    restricted_bids.append(all_bids[current_bid_index])
                    current_bid_index += 1
                    bids_removed += 1
            del fig

    def plotBinProfiles(self, timer):
        """Plot the distributions of kmer and coverage signatures"""
        self.BM.loadBins(timer,
                         makeBins=True,
                         silent=False,
                         bids=self.bids,
                         transform=self.transform)
        if len(self.BM.bins) == 0:
            print "Sorry, no bins to plot"
        else:
            print "Plotting bin profiles"
            self.BM.setColorMap(self.cmString)
            self.BM.plotProfileDistributions()

    def plotContigs(self, timer, coreCut, all=False):
        """plot contigs"""
        if all:
            print "Plotting all contigs"
            self.PM.plotAll(timer, coreCut, transform=self.transform, ignoreContigLengths=self.ignoreContigLengths)
        else:
            self.BM.loadBins(timer,
                             makeBins=True,
                             silent=False,
                             bids=self.bids,
                             transform=self.transform)
            if len(self.BM.bins) == 0:
                print "Sorry, no bins to plot"
            else:
                print "Plotting binned contigs"
                self.BM.setColorMap(self.cmString)
                if self.bids == []:
                    self.bids = self.BM.getBids()
                self.BM.plotMultipleBins([self.bids], squash=True, ignoreContigLengths=self.ignoreContigLengths)

    def plotBinAssignents(self, timer, coreCut):
        """visualise bin assignments"""
        self.BM.loadBins(timer,
                         makeBins=True,
                         getUnbinned=True,
                         silent=False,
                         bids=self.bids,
                         cutOff=coreCut,
                         transform=self.transform)
        if len(self.BM.bins) == 0:
            print "Sorry, no bins to plot"
        else:
            print "Plotting bin assignments"
            if self.bids == []:
                self.bids = self.BM.getBids()

            # work out colors per bin
            S = 1.0
            V = 1.0
            num_bins = len(self.bids)
            offset = 0.5
            step = 1. /(num_bins-1 + 2. * offset)
            Hs = np.array([step*(i + offset) for i in range(num_bins)])
            cols = [htr(H, S, V) for H in Hs]
            np.random.shuffle(cols)
            unbinned_col = (0.2,0.2,0.2)
            bin_cols = {0:(0.2,0.2,0.2)}    # unassigned color
            i = 0
            for bid in self.bids:
                bin_cols[bid] = cols[i]
                i += 1

            # for binned contigs
            b_disp_vals = []
            b_disp_cols = []
            b_disp_lens = []
            b_num = 0
            # for the unwashed masses
            ub_disp_vals = []
            ub_disp_cols = []
            ub_disp_lens = []
            ub_num = 0
                
            # assign a color to each contig
            for row_index in np.arange(len(self.PM.indices)):
                if self.PM.binIds[row_index] == 0:
                    ub_disp_cols.append(unbinned_col)
                    ub_disp_vals.append(self.PM.transformedCP[row_index])
                    ub_disp_lens.append(self.PM.contigLengths[row_index])
                    ub_num += 1
                else:
                    b_disp_cols.append(bin_cols[self.PM.binIds[row_index]])
                    b_disp_vals.append(self.PM.transformedCP[row_index])
                    b_disp_lens.append(self.PM.contigLengths[row_index])
                    b_num += 1

            b_disp_vals = np.reshape(b_disp_vals, (b_num, 3))
            fig = plt.figure()
            #import matplotlib.gridspec as gridspec
            #gs = gridspec.GridSpec(2,1,height_ratios=[4,1])
            #ax1 = plt.subplot(gs[0], projection='3d')
            ax1 = plt.subplot(1,1,1, projection='3d')
            
            if ub_num != 0:
                ub_disp_vals = np.reshape(ub_disp_vals, (ub_num, 3))
                if self.ignoreContigLengths:
                    sc = ax1.scatter(ub_disp_vals[:,0],
                                     ub_disp_vals[:,1],
                                     ub_disp_vals[:,2],
                                     edgecolors='none',
                                     c=ub_disp_cols,
                                     s=10.,
                                     marker='.',
                                     alpha=0.35)
                else:
                    sc = ax1.scatter(ub_disp_vals[:,0],
                                     ub_disp_vals[:,1],
                                     ub_disp_vals[:,2],
                                     edgecolors='none',
                                     c=ub_disp_cols,
                                     s=np.sqrt(ub_disp_lens),
                                     marker='.',
                                     alpha=0.35)

            if self.ignoreContigLengths:
                sc = ax1.scatter(b_disp_vals[:,0],
                                 b_disp_vals[:,1],
                                 b_disp_vals[:,2],
                                 edgecolors='none',
                                 c=b_disp_cols,
                                 s=10.,
                                 marker='.')
            else:
                sc = ax1.scatter(b_disp_vals[:,0],
                                 b_disp_vals[:,1],
                                 b_disp_vals[:,2],
                                 edgecolors='none',
                                 c=b_disp_cols,
                                 s=np.sqrt(b_disp_lens),
                                 marker='.')

            (bin_centroid_points, bin_centroid_colors, bin_centroid_gc, bin_ids) = self.BM.findCoreCentres(processChimeric=True)
            outer_index = 0
            for bid in self.bids:
                bbox_props = dict(boxstyle="round4", fc="w", ec=bin_cols[bid], alpha=0.6)
                ax1.text(bin_centroid_points[outer_index, 0],
                         bin_centroid_points[outer_index, 1],
                         bin_centroid_points[outer_index, 2],
                         str(int(bid)),
                         color='black',
                         size='large',
                         ha="center", va="center", bbox=bbox_props
                         )
                outer_index += 1
            ax1.set_xticklabels([])
            ax1.set_yticklabels([])
            ax1.set_zticklabels([])
            ax1.set_xticks([])
            ax1.set_yticks([])
            ax1.set_zticks([])
            
            plt.title(self.PM.dbFileName)
            
            if False:
                #plot bin keys
                ax2 = plt.subplot(gs[1])
                
                # work out text placement
                bin_rows = int(np.sqrt(num_bins)) + 1
                bin_columns = int(float(num_bins)/float(bin_rows)) + 2
                bin_rows += 1 
                ax2.set_xlim(0, bin_columns)
                ax2.set_ylim(0, bin_rows)
                bid_coords = zip(np.array([[i]*bin_columns for i in range(bin_rows)]).flatten(), range(bin_columns) * bin_rows)
    
                outer_index = 0
                
                for bid in self.bids:
                    bbox_props = dict(boxstyle="round", fc=bin_cols[bid])
                    ax2.text(bid_coords[outer_index][0] +1,
                             bid_coords[outer_index][1] +1,
                             str(int(bid)),
                             ha="center", va="center",
                             bbox=bbox_props,
                             size='medium' 
                             )
                    outer_index += 1
    
                plt.xticks([], [])
                plt.yticks([], [])

            plt.show()
            plt.close(fig)
            del fig


    def plotPoints(self, timer):
        """plot points"""
        self.BM.loadBins(timer,
                         makeBins=True,
                         silent=False,
                         bids=self.bids,
                         transform=self.transform)
        if len(self.BM.bins) == 0:
            print "Sorry, no bins to plot"
        else:
            print "Plotting bin points"
            self.BM.setColorMap(self.cmString)
            self.BM.plotBinPoints()

    def plotCompare(self, timer, coreCut):
        """Plot cores side by side with their contigs"""
        self.PM2 = binManager.ProfileManager(dbFileName=self.BM.PM.dbFileName)
        self.PM2.loadData(timer,
                          condition="length >= "+str(coreCut),
                          bids=self.bids,
                          loadContigNames=False,
                          loadContigLengths=True,
                          )
        if self.transform:
            self.PM2.transformCP(timer)

        self.BM.loadBins(timer,
                         makeBins=True,
                         loadContigNames=False,
                         bids=self.bids,
                         silent=False,
                         transform=self.transform)

        self.PM2.setColorMap(self.cmString)
        self.BM.setColorMap(self.cmString)
        
        if len(self.BM.bins) == 0:
            print "Sorry, no bins to plot"
        else:
            print "Plotting side by side graphs"
            (bin_centroid_points, bin_centroid_colors, bin_centroid_gc, bin_ids) = self.BM.findCoreCentres()
            self.plotCoresVsContigs(bin_centroid_points, bin_centroid_colors)

    def plotIds(self, timer):
        """Make a 3d plot of the bins but use IDs instead of points

        This function will help users know which bins to merge
        """
        self.BM.loadBins(timer,
                         makeBins=True,
                         silent=False,
                         bids=self.bids,
                         transform=self.transform)
        if len(self.BM.bins) == 0:
            print "Sorry, no bins to plot"
        else:
            print "Plotting bin IDs"
            self.BM.setColorMap(self.cmString)
            self.BM.plotBinIds()

    def plotUnbinned(self, timer, coreCut):
        """Plot all contigs over a certain length which are unbinned"""
        print "Plotting unbinned contigs"
        self.PM.plotUnbinned(timer, coreCut, transform=self.transform, ignoreContigLengths=self.ignoreContigLengths)

    def plotSideBySide(self, timer, coreCut):
        """Plot all bins separately in one large image"""
        self.BM.loadBins(timer,
                         makeBins=True,
                         silent=False,
                         bids=self.bids,
                         transform=self.transform)
        if len(self.BM.bins) == 0:
            print "Sorry, no bins to plot"
        else:
            self.BM.setColorMap(self.cmString)
            self.BM.plotBins(sideBySide=True,
                             plotEllipsoid=True,
                             ignoreContigLengths=self.ignoreContigLengths)

    def plotTogether(self, timer, coreCut, doMers=False):
        """Plot all bins in ellipses on one normal image"""
        self.BM.loadBins(timer,
                         makeBins=True,
                         silent=False,
                         bids=self.bids,
                         transform=self.transform)
        if len(self.BM.bins) == 0:
            print "Sorry, no bins to plot"
        else:
            print "Plotting all bins together"
            self.BM.setColorMap(self.cmString)
            if self.bids == []:
                p_bids = self.BM.getBids()
            else:
                p_bids = self.bids
            self.BM.plotSelectBins(p_bids,
                                   plotMers=doMers,
                                   plotEllipsoid=True,
                                   ignoreContigLengths=self.ignoreContigLengths)
         

#------------------------------------------------------------------------------
# IO and IMAGE RENDERING

    def plotCoresVsContigs(self, binCentroidPoints, binCentroidColors, azim=0, elev=0, fileName='', dpi=300, format='png'):
        """Render the image for validating cores"""
        if(fileName==""):
            # plot on screen for user
            fig = plt.figure()
            ax1 = fig.add_subplot(121, projection='3d')
            if self.ignoreContigLengths:
                sc = ax1.scatter(self.PM2.transformedCP[:,0], self.PM2.transformedCP[:,1], self.PM2.transformedCP[:,2], edgecolors='none', c=self.PM2.contigGCs, cmap=self.PM2.colorMapGC, vmin=0.0, vmax=1.0, s=10, marker='.')
            else:
                sc = ax1.scatter(self.PM2.transformedCP[:,0], self.PM2.transformedCP[:,1], self.PM2.transformedCP[:,2], edgecolors='k', c=self.PM2.contigGCs, cmap=self.PM2.colorMapGC, vmin=0.0, vmax=1.0, s=np.sqrt(self.PM2.contigLengths), marker='.')                
            sc.set_edgecolors = sc.set_facecolors = lambda *args:None # disable depth transparency effect

            ax2 = fig.add_subplot(122, projection='3d')
            sc = ax2.scatter(binCentroidPoints[:,0], binCentroidPoints[:,1], binCentroidPoints[:,2], edgecolors=binCentroidColors, c=binCentroidColors)
            sc.set_edgecolors = sc.set_facecolors = lambda *args:None # disable depth transparency effect

            ax2.set_xlim(ax1.get_xlim())
            ax2.set_ylim(ax1.get_ylim())
            ax2.set_zlim(ax1.get_zlim())


            self.BM.plotStoitNames(ax1)
            self.BM.plotStoitNames(ax2)
            try:
                plt.show()
                plt.close(fig)
            except:
                print "Error showing image", sys.exc_info()[0]
                raise
            del fig
        else:
            f_name1 = fileName + "_1"
            f_name2 = fileName + "_2"
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            ax.scatter(self.PM2.transformedCP[:,0], self.PM2.transformedCP[:,1], self.PM2.transformedCP[:,2], edgecolors='none', c=self.PM2.contigGCs, cmap=self.PM2.colorMapGC, vmin=0.0, vmax=1.0, marker='.')
            ax.azim = azim
            ax.elev = elev
            ax.set_xlim3d(0,self.PM2.scaleFactor)
            ax.set_ylim3d(0,self.PM2.scaleFactor)
            ax.set_zlim3d(0,self.PM2.scaleFactor)
            ax.set_xticklabels([])
            ax.set_yticklabels([])
            ax.set_zticklabels([])
            ax.set_xticks([])
            ax.set_yticks([])
            ax.set_zticks([])
            self.BM.plotStoitNames(ax)

            try:
                fig.set_size_inches(12,12)
                plt.savefig(f_name1,dpi=dpi,format=format)
                plt.close(fig)
            except:
                print "Error saving image",f_name1, sys.exc_info()[0]
                raise
            del fig

            fig = plt.figure()
            ax = fig.gca(projection='3d')
            outer_index = 0
            for bid in self.BM.getBids():
                ax.text(binCentroidPoints[outer_index,0],
                        binCentroidPoints[outer_index,1],
                        binCentroidPoints[outer_index,2],
                        str(int(bid)),
                        color=binCentroidColors[outer_index]
                        )
                outer_index += 1

            ax.azim = azim
            ax.elev = elev
            ax.set_xlim3d(0,self.PM2.scaleFactor)
            ax.set_ylim3d(0,self.PM2.scaleFactor)
            ax.set_zlim3d(0,self.PM2.scaleFactor)
            ax.set_xticklabels([])
            ax.set_yticklabels([])
            ax.set_zticklabels([])
            ax.set_xticks([])
            ax.set_yticks([])
            ax.set_zticks([])
            self.BM.plotStoitNames(ax)

            try:
                fig.set_size_inches(12,12)
                plt.savefig(f_name2,dpi=dpi,format=format)
                plt.close(fig)
            except:
                print "Error saving image",f_name1, sys.exc_info()[0]
                raise
            del fig

###############################################################################
###############################################################################
###############################################################################
###############################################################################
